"""
transaction_set.py

Defines the Professional Claims 837 005010X222A2 transaction set model.
"""

from x12.models import X12SegmentGroup
from .loops import Header, Footer, Loop1000A, Loop1000B, Loop2000A, Loop2300
from pydantic import Field, root_validator
from typing import List, Dict
from x12.validators import validate_segment_count


class HealthCareClaimProfessional(X12SegmentGroup):
    """
    The HealthCare Claim - Professional transaction model (837)
    """

    header: Header
    loop_1000a: Loop1000A
    loop_1000b: Loop1000B
    loop_2000a: List[Loop2000A] = Field(min_items=1)
    footer: Footer

    _validate_segment_count = root_validator(allow_reuse=True)(validate_segment_count)

    @property
    def patient(self, return_first=True):
        """
        Returns patient demographics for the health care claim
        """
        result = []
        for billing_provider in self.loop_2000a:
            for subscriber in billing_provider.loop_2000b:
                if subscriber.hl_segment.hierarchical_child_code == "0":
                    result.append(
                        subscriber.dict(
                            exclude={"loop_2010bb", "loop_2300", "loop_2000c"}
                        )
                    )
                else:
                    for dependent in subscriber.loop_2000c:
                        result.append(dependent.dict(exclude={"loop_2300"}))

        return result[0] if return_first else result

    def _parse_claim_data(self, loop_2300: Loop2300) -> Dict:
        """Returns the claim"""

        claim_data = {"clm_segment": loop_2300.clm_segment.dict(), "sv1_segment": []}

        for service_line in loop_2300.loop_2400:
            claim_data["sv1_segment"].append(service_line.sv1_segment)

        return claim_data

    @property
    def claim(self, return_first=True):
        """
        Returns the header and line items for the health care claim
        """
        result = []
        for billing_provider in self.loop_2000a:
            for subscriber in billing_provider.loop_2000b:
                if subscriber.hl_segment.hierarchical_child_code == "0":
                    for claim in subscriber.loop_2300:
                        result.append(self._parse_claim_data(claim))
                else:
                    for dependent in subscriber.loop_2000c:
                        for claim in dependent.loop_2300:
                            result.append(self._parse_claim_data(claim))

        return result[0] if return_first else result



"""
transaction_set.py

Defines the Eligibility 271 005010X279A1 transaction set model.
"""

from typing import List, Dict, Union, Tuple

from x12.models import X12SegmentGroup

from .loops import Footer, Header, Loop2000A
from pydantic import root_validator
from x12.validators import validate_segment_count


class EligibilityBenefit(X12SegmentGroup):
    """
    The ASC X12 271 (EligibilityBenefit) transaction model.
    """

    header: Header
    loop_2000a: List[Loop2000A]
    footer: Footer

    _validate_segment_count = root_validator(allow_reuse=True)(validate_segment_count)

    @root_validator
    def validate_subscriber_name(cls, values):
        """
        Validates that the subscriber mame is present if the subscriber is a patient

        :param values: The raw, unvalidated transaction data.
        """
        for info_source in values.get("loop_2000a", []):
            for info_receiver in info_source.loop_2000b:
                for subscriber in info_receiver.loop_2000c:
                    child_code = subscriber.hl_segment.hierarchical_child_code
                    first_name = subscriber.loop_2100c.nm1_segment.name_first

                    if child_code == "0" and not first_name:
                        raise ValueError(
                            f"name_first is required when the subscriber is the patient"
                        )

        return values

    @root_validator
    def validate_subscriber_hierarchy_child_code(cls, values):
        """
        Validates that a subscriber's hierarchy child code is set correctly based on the presence of a dependent loop.

        :param values: The raw, unvalidated transaction data.
        """
        for info_source in values.get("loop_2000a", []):
            for info_receiver in info_source.loop_2000b:
                for subscriber in info_receiver.loop_2000c:
                    child_code = subscriber.hl_segment.hierarchical_child_code

                    if child_code == "1" and not subscriber.loop_2000d:
                        raise ValueError(
                            f"Invalid subscriber hierarchy code {child_code} no dependent record is present"
                        )
        return values

    @root_validator
    def validate_hierarchy_ids(cls, values):
        """
        Validates the HL segments linkage in regards to the entire EligibilityInquiry transaction.
        Validations are limited to checks that are not covered within a segment or field scope.

        :param values: The raw, unvalidated transaction data.
        """

        def get_ids(hl_segment: Dict) -> Tuple[int, int]:
            """returns tuple of (id, parent_id)"""
            id = hl_segment.hierarchical_id_number
            parent_id = hl_segment.hierarchical_parent_id_number
            return int(id) if id else 0, int(parent_id) if parent_id else 0

        for info_source in values.get("loop_2000a", []):
            # info source does not have a parent id, since it starts a new hierarchy - this is validated at the
            # segment level
            source_id, _ = get_ids(info_source.hl_segment)
            previous_id: int = source_id

            for info_receiver in info_source.loop_2000b:
                receiver_id, receiver_parent_id = get_ids(info_receiver.hl_segment)

                if receiver_parent_id != previous_id:
                    raise ValueError(f"Invalid receiver parent id {receiver_parent_id}")

                if receiver_parent_id != source_id:
                    raise ValueError(
                        f"receiver parent id {receiver_parent_id} != source id {source_id}"
                    )

                previous_id = receiver_id

                for subscriber in info_receiver.loop_2000c:
                    subscriber_id, subscriber_parent_id = get_ids(subscriber.hl_segment)

                    if subscriber_parent_id != previous_id:
                        raise ValueError(
                            f"Invalid subscriber parent id {subscriber_parent_id}"
                        )

                    if subscriber_parent_id != receiver_id:
                        raise ValueError(
                            f"subscriber parent id {subscriber_parent_id} != receiver id {receiver_id}"
                        )

                    previous_id = subscriber_id

                    if not subscriber.loop_2000d:
                        continue

                    for dependent in subscriber.loop_2000d:
                        dependent_id, dependent_parent_id = get_ids(
                            dependent.hl_segment
                        )

                        if dependent_parent_id != previous_id:
                            raise ValueError(
                                f"Invalid dependent parent id {dependent_parent_id}"
                            )

                        if dependent_parent_id != subscriber_id:
                            raise ValueError(
                                f"dependent parent id {dependent_parent_id} != subscriber id {subscriber_id}"
                            )

                        previous_id = dependent_id
        return values

    @property
    def information_source(self, return_first=True) -> Union[List, Dict]:
        """
        Returns Information Source Loops 2000A and 2100A within a single "record" (dictionary).

        :param return_first: Indicates if the first result is returned. When set to False all results are returned.
        :returns: Dictionary if return_first = True, otherwise returns a List
        """
        result = [
            info_source.dict(exclude={"loop_2000b"}) for info_source in self.loop_2000a
        ]
        return result[0] if return_first else result

    @property
    def information_receiver(self, return_first=True) -> Union[List, Dict]:
        """
        Returns Information Receiver Loops 2000B and 2100B within a single "record" (dictionary).

        :param return_first: Indicates if the first result is returned. When set to False all results are returned.
        :returns: Dictionary if return_first = True, otherwise returns a List
        """
        result = []

        for info_source in self.loop_2000a:
            for info_receiver in info_source.loop_2000b:
                result.append(info_receiver.dict(exclude={"loop_2000c"}))
        return result[0] if return_first else result

    @property
    def subscriber(self, return_first=True) -> Union[List, Dict]:
        """
        Returns Subscriber Loops 2000C, 2100C, and 2110C within a single "record" (dictionary).

        :param return_first: Indicates if the first result is returned. When set to False all results are returned.
        :returns: Dictionary if return_first = True, otherwise returns a List
        """
        result = []

        for info_source in self.loop_2000a:
            for info_receiver in info_source.loop_2000b:
                for subscriber in info_receiver.loop_2000c:
                    result.append(subscriber.dict(exclude={"loop_2000d"}))
        return result[0] if return_first else result

    @property
    def dependent(self, return_first=True) -> Union[List, Dict]:
        """
        Returns Dependent Loops 2000D, 2100D, and 2110D within a single "record" (dictionary).

        :param return_first: Indicates if the first result is returned. When set to False all results are returned.
        :returns: Dictionary or List if a dependent is present, based on `return_first`. If no dependent is present, returns None.
        """
        result = []

        for info_source in self.loop_2000a:
            for info_receiver in info_source.loop_2000b:
                for subscriber in info_receiver.loop_2000c:
                    if subscriber.loop_2000d:
                        for dependent in subscriber.loop_2000d:
                            result.append(dependent.dict())

        if return_first:
            return result[0] if result else None
        else:
            return result if len(result) else None



"""
test_x12_model_reader.py

Supports general model streaming tests validating the number of models returned, expected payload, etc.
"""
from x12.io import X12ModelReader


def test_multiple_transactions(
    x12_control_header, x12_control_footer, x12_270_subscriber_transaction
):
    transactions: str = x12_270_subscriber_transaction * 2
    x12_input = f"{x12_control_header}{transactions}{x12_control_footer}"

    with X12ModelReader(x12_input) as r:
        model_result = [m for m in r.models()]
        assert len(model_result) == 2
        assert model_result[0].x12() == x12_270_subscriber_transaction


def test_custom_delimiters(x12_with_custom_delimiters):

    # remove control segments from the x12 transaction with custom delimiters
    control_segments = ("ISA", "GS", "GE", "IEA")
    expected_segments = []

    for x in x12_with_custom_delimiters.split("\n"):
        if x.split("|")[0] in control_segments:
            continue
        expected_segments.append(x)

    expected_transaction = "\n".join(expected_segments)

    with X12ModelReader(x12_with_custom_delimiters) as r:
        model_result = [m for m in r.models()]
        assert len(model_result) == 1
        assert model_result[0].x12() == expected_transaction



"""
parsing.py

Parses X12 271 005010X279A1 segments into a transactional domain model.

The parsing module includes a specific parser for the 271 transaction and loop parsing functions to create new loops
as segments are streamed to the transactional data model.

Loop parsing functions are implemented as set_[description]_loop(context: X12ParserContext, segment_data: Dict).
"""
from enum import Enum

from x12.parsing import match, X12ParserContext
from typing import Dict, Set


class TransactionLoops(str, Enum):
    """
    The loops used to support the 271 005010X279A1 format.
    """

    HEADER = "header"
    INFORMATION_SOURCE = "loop_2000a"
    INFORMATION_SOURCE_NAME = "loop_2100a"
    INFORMATION_RECEIVER = "loop_2000b"
    INFORMATION_RECEIVER_NAME = "loop_2100b"
    SUBSCRIBER = "loop_2000c"
    SUBSCRIBER_NAME = "loop_2100c"
    SUBSCRIBER_ELIGIBILITY = "loop_2110c"
    SUBSCRIBER_ADDITIONAL_ELIGIBILITY = "loop_2115c"
    SUBSCRIBER_RELATED_BENEFIT_NAME = "loop_2120c"
    DEPENDENT = "loop_2000d"
    DEPENDENT_NAME = "loop_2100d"
    DEPENDENT_ELIGIBILITY = "loop_2110d"
    DEPENDENT_ADDITIONAL_ELIGIBILITY = "loop_2115d"
    DEPENDENT_RELATED_BENEFIT_NAME = "loop_2120d"
    FOOTER = "footer"


SUBSCRIBER_LOOPS: Set[TransactionLoops] = {
    TransactionLoops.SUBSCRIBER,
    TransactionLoops.SUBSCRIBER_NAME,
    TransactionLoops.SUBSCRIBER_ELIGIBILITY,
    TransactionLoops.SUBSCRIBER_ADDITIONAL_ELIGIBILITY,
    TransactionLoops.SUBSCRIBER_RELATED_BENEFIT_NAME,
}

DEPENDENT_LOOPS: Set[TransactionLoops] = {
    TransactionLoops.DEPENDENT,
    TransactionLoops.DEPENDENT_NAME,
    TransactionLoops.DEPENDENT_ELIGIBILITY,
    TransactionLoops.DEPENDENT_ADDITIONAL_ELIGIBILITY,
    TransactionLoops.DEPENDENT_RELATED_BENEFIT_NAME,
}


def _get_info_source(context) -> Dict:
    """Returns the current information source"""

    return context.transaction_data[TransactionLoops.INFORMATION_SOURCE][-1]


def _get_info_receiver(context) -> Dict:
    """Returns the current information receiver"""

    info_source = _get_info_source(context)
    return info_source[TransactionLoops.INFORMATION_RECEIVER][-1]


def _get_subscriber(context) -> Dict:
    """Returns the current subscriber"""
    info_receiver = _get_info_receiver(context)
    return info_receiver[TransactionLoops.SUBSCRIBER][-1]


def _get_dependent(context) -> Dict:
    """Returns the current dependent"""
    subscriber = _get_subscriber(context)
    return subscriber[TransactionLoops.DEPENDENT][-1]


def _get_subscriber_eligibility(context) -> Dict:
    """Returns the current subscriber eligibility"""
    subscriber = _get_subscriber(context)
    return subscriber[TransactionLoops.SUBSCRIBER_NAME][
        TransactionLoops.SUBSCRIBER_ELIGIBILITY
    ][-1]


def _get_dependent_eligibility(context) -> Dict:
    """Returns the current dependent eligibility"""
    dependent = _get_dependent(context)
    return dependent[TransactionLoops.DEPENDENT_NAME][
        TransactionLoops.DEPENDENT_ELIGIBILITY
    ][-1]


@match("ST")
def set_header_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the transaction set header loop.
    Creates initial data structures for the record

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    context.transaction_data[TransactionLoops.HEADER] = {"bht_segment": None}

    context.set_loop_context(
        TransactionLoops.HEADER, context.transaction_data[TransactionLoops.HEADER]
    )


@match("HL", {"hierarchical_level_code": "20"})
def set_info_source_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the information source loop (Loop 2000A).

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    info_source = {
        "hl_segment": None,
        "aaa_segment": [],
        "loop_2100a": None,
        "loop_2000b": [],
    }

    if TransactionLoops.INFORMATION_SOURCE not in context.transaction_data:
        context.transaction_data[TransactionLoops.INFORMATION_SOURCE] = []

    context.transaction_data[TransactionLoops.INFORMATION_SOURCE].append(info_source)
    current_record = context.transaction_data[TransactionLoops.INFORMATION_SOURCE][-1]
    context.set_loop_context(TransactionLoops.INFORMATION_SOURCE, current_record)


@match("NM1")
def set_info_source_name_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the information source name loop (Loop 2100A)

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    if context.loop_name != TransactionLoops.INFORMATION_SOURCE:
        return

    info_source_name = {"nm1_segment": None, "per_segment": [], "aaa_segment": []}

    info_source = _get_info_source(context)
    info_source[TransactionLoops.INFORMATION_SOURCE_NAME] = info_source_name
    context.set_loop_context(TransactionLoops.INFORMATION_SOURCE_NAME, info_source_name)


@match("HL", {"hierarchical_level_code": "21"})
def set_info_receiver_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the information receiver loop (Loop 2000B)

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    info_source = _get_info_source(context)

    info_source[TransactionLoops.INFORMATION_RECEIVER].append(
        {"hl_segment": None, "loop_2100b": None, "loop_2000c": []}
    )

    info_receiver = info_source[TransactionLoops.INFORMATION_RECEIVER][-1]
    context.set_loop_context(TransactionLoops.INFORMATION_RECEIVER, info_receiver)


@match("NM1")
def set_info_receiver_name_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the information receiver name loop (Loop 2100B)

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    if context.loop_name != TransactionLoops.INFORMATION_RECEIVER:
        return

    info_receiver = _get_info_receiver(context)
    info_receiver[TransactionLoops.INFORMATION_RECEIVER_NAME] = {
        "nm1_segment": None,
        "ref_segment": [],
        "n3_segment": None,
        "n4_segment": None,
        "aaa_segment": None,
        "prv_segment": None,
    }

    info_receiver_name = info_receiver[TransactionLoops.INFORMATION_RECEIVER_NAME]
    context.set_loop_context(
        TransactionLoops.INFORMATION_RECEIVER_NAME, info_receiver_name
    )


@match("HL", {"hierarchical_level_code": "22"})
def set_subscriber_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the subscriber loop (Loop 2000C)

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    info_receiver = _get_info_receiver(context)

    info_receiver[TransactionLoops.SUBSCRIBER].append(
        {"hl_segment": None, "trn_segment": [], "loop_2100c": None, "loop_2000d": []}
    )

    subscriber = info_receiver[TransactionLoops.SUBSCRIBER][-1]
    context.set_loop_context(TransactionLoops.SUBSCRIBER, subscriber)


@match("HL", {"hierarchical_level_code": "23"})
def set_dependent_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the dependent loop (Loop 2000D)

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    subscriber = _get_subscriber(context)

    subscriber[TransactionLoops.DEPENDENT].append(
        {
            "hl_segment": None,
            "trn_segment": [],
            "loop_2100d": None,
        }
    )

    dependent = subscriber[TransactionLoops.DEPENDENT][-1]
    context.set_loop_context(TransactionLoops.DEPENDENT, dependent)


@match("NM1")
def set_member_name_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the subscriber or dependent name loop (Loop 2100C or 2100D)

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    if context.loop_name not in (
        TransactionLoops.SUBSCRIBER,
        TransactionLoops.DEPENDENT,
    ):
        return

    if context.loop_name == TransactionLoops.SUBSCRIBER:
        member = _get_subscriber(context)
        loop_name = TransactionLoops.SUBSCRIBER_NAME
        eligibility_loop = TransactionLoops.SUBSCRIBER_ELIGIBILITY
    else:
        member = _get_dependent(context)
        loop_name = TransactionLoops.DEPENDENT_NAME
        eligibility_loop = TransactionLoops.DEPENDENT_ELIGIBILITY

    member[loop_name] = {
        "nm1_segment": None,
        "ref_segment": [],
        "n3_segment": None,
        "n4_segment": None,
        "aaa_segment": [],
        "prv_segment": None,
        "dmg_segment": None,
        "ins_segment": None,
        "hi_segment": None,
        "dtp_segment": [],
        "mpi_segment": None,
        eligibility_loop: [],
    }

    context.set_loop_context(loop_name, member[loop_name])


@match("EB")
def set_member_eligibility_benefit_info_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Sets the subscriber or dependent eligibility/benefit info loop (Loop 2110C/2110D)

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    if context.loop_name not in (
        TransactionLoops.SUBSCRIBER_NAME,
        TransactionLoops.SUBSCRIBER_ELIGIBILITY,
        TransactionLoops.DEPENDENT_NAME,
        TransactionLoops.DEPENDENT_ELIGIBILITY,
    ):
        return

    if context.loop_name in SUBSCRIBER_LOOPS:
        member = _get_subscriber(context)
        record = member[TransactionLoops.SUBSCRIBER_NAME][
            TransactionLoops.SUBSCRIBER_ELIGIBILITY
        ]
        eligibility_loop = TransactionLoops.SUBSCRIBER_ELIGIBILITY
        additional_info_loop = TransactionLoops.SUBSCRIBER_ADDITIONAL_ELIGIBILITY
        benefit_entity_loop = TransactionLoops.SUBSCRIBER_RELATED_BENEFIT_NAME
    else:
        member = _get_dependent(context)
        record = member[TransactionLoops.DEPENDENT_NAME][
            TransactionLoops.DEPENDENT_ELIGIBILITY
        ]
        eligibility_loop = TransactionLoops.DEPENDENT_ELIGIBILITY
        additional_info_loop = TransactionLoops.DEPENDENT_ADDITIONAL_ELIGIBILITY
        benefit_entity_loop = TransactionLoops.DEPENDENT_RELATED_BENEFIT_NAME

    record.append(
        {
            "eb_segment": None,
            "hsd_segment": [],
            "ref_segment": [],
            "dtp_segment": [],
            "aaa_segment": [],
            "msg_segment": [],
            additional_info_loop: [],
            "ls_segment": None,
            benefit_entity_loop: [],
            "le_segment": None,
        }
    )

    member_eligibility = record[-1]
    context.set_loop_context(eligibility_loop, member_eligibility)


@match("III")
def set_member_additional_information(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Sets the subscriber/dependent eligibility benefit additional information loop (Loop 2115c/2115d)

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    if context.loop_name in SUBSCRIBER_LOOPS:
        eligibility = _get_subscriber_eligibility(context)
        loop_name = TransactionLoops.SUBSCRIBER_ADDITIONAL_ELIGIBILITY
    else:
        eligibility = _get_dependent_eligibility(context)
        loop_name = TransactionLoops.DEPENDENT_ADDITIONAL_ELIGIBILITY

    eligibility[loop_name].append({"iii_segment": None})

    additional_info = eligibility[loop_name][-1]
    context.set_loop_context(loop_name, additional_info)


@match("LS")
def set_loop_header_in_eligibility_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Ensures that the Loop Header Segment is written to the Member (Subscriber or Dependent)

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    if context.loop_name in SUBSCRIBER_LOOPS:
        related_benefit = TransactionLoops.SUBSCRIBER_RELATED_BENEFIT_NAME
        loop_name = TransactionLoops.SUBSCRIBER_ELIGIBILITY
        eligibility_record = _get_subscriber_eligibility(context)
    else:
        related_benefit = TransactionLoops.DEPENDENT_RELATED_BENEFIT_NAME
        loop_name = TransactionLoops.DEPENDENT_ELIGIBILITY
        eligibility_record = _get_dependent_eligibility(context)

    eligibility_record[related_benefit].append(
        {
            "nm1_segment": None,
            "n3_segment": None,
            "n4_segment": None,
            "per_segment": [],
            "prv_segment": None,
        }
    )

    context.set_loop_context(loop_name, eligibility_record)


@match("NM1")
def set_benefit_related_entity_name_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Creates the structure for the Benefit Related Entity Name Loop.
    The Benefit Related Entity Name Loop is entered either through the Eligibility or Additional Eligibility Loop.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    if context.loop_name not in (
        TransactionLoops.SUBSCRIBER_ELIGIBILITY,
        TransactionLoops.SUBSCRIBER_ADDITIONAL_ELIGIBILITY,
        TransactionLoops.DEPENDENT_ELIGIBILITY,
        TransactionLoops.DEPENDENT_ADDITIONAL_ELIGIBILITY,
    ):
        return

    if context.loop_name in SUBSCRIBER_LOOPS:
        eligibility = _get_subscriber_eligibility(context)
        loop_name = TransactionLoops.SUBSCRIBER_RELATED_BENEFIT_NAME
    else:
        eligibility = _get_dependent_eligibility(context)
        loop_name = TransactionLoops.DEPENDENT_RELATED_BENEFIT_NAME

    related_entity = eligibility[loop_name][-1]
    context.set_loop_context(loop_name, related_entity)


@match("LE")
def set_loop_footer_in_eligibility_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Ensures that the Loop Header Segment is written to the Member (Subscriber or Dependent)

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    if context.loop_name in SUBSCRIBER_LOOPS:
        eligibility = _get_subscriber_eligibility(context)
        context.set_loop_context(TransactionLoops.SUBSCRIBER_ELIGIBILITY, eligibility)
    else:
        eligibility = _get_dependent_eligibility(context)
        context.set_loop_context(TransactionLoops.DEPENDENT_ELIGIBILITY, eligibility)


@match("SE")
def set_se_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the transaction set footer loop.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """

    context.set_loop_context(
        TransactionLoops.FOOTER, context.transaction_data["footer"]
    )



"""
Tests LinuxForHealth X12SegmentReader
"""
import pytest

import x12.io
from x12.io import X12SegmentReader


@pytest.mark.parametrize(
    "test_input", ["simple_270_with_new_lines", "simple_270_one_line"]
)
def test_init(request, test_input: str):
    """
    Tests X12SegmentReader initialization

    :param request: The pytest request fixture. Used to lookup fixture values by name.
    :param test_input: The fixture name.
    """

    input_value = request.getfixturevalue(test_input)
    x12_reader = X12SegmentReader(input_value)
    assert x12_reader._x12_input
    assert x12_reader._buffer_size is None
    assert x12_reader._x12_stream is None
    assert x12_reader.delimiters is None


@pytest.mark.parametrize(
    "test_input", ["simple_270_with_new_lines", "simple_270_one_line"]
)
def test_segments_with_string_data(request, test_input: str):
    """
    Test X12SegmentReader models with a x12 data/message input.

    :param request: The pytest request fixture. Used to lookup fixture values by name.
    :param test_input: The fixture name.
    """

    input_value = request.getfixturevalue(test_input)

    with X12SegmentReader(input_value) as r:
        segment_count = 0

        assert r.delimiters.component_separator == ":"
        assert r.delimiters.element_separator == "*"
        assert r.delimiters.repetition_separator == "^"
        assert r.delimiters.segment_terminator == "~"

        for _ in r.segments():
            segment_count += 1

        assert 21 == segment_count

    assert r._x12_stream.closed
    assert r.delimiters is None
    assert r._x12_input is None


@pytest.mark.parametrize(
    "test_input", ["simple_270_with_new_lines", "simple_270_one_line"]
)
def test_segments_with_file_path(request, tmpdir, test_input: str):
    """
    Test X12SegmentReader models with a x12 file path input.

    :param request: The pytest request fixture. Used to lookup fixture values by name.
    :param tmpdir: Pytest fixture used to create temporary files and directories
    :param test_input: The fixture name.
    """

    input_value = request.getfixturevalue(test_input)
    f = tmpdir.mkdir("x12-support").join("test.x12")
    f.write(input_value)

    with X12SegmentReader(f) as r:
        segment_count = 0
        version_key = None

        assert r.delimiters.component_separator == ":"
        assert r.delimiters.element_separator == "*"
        assert r.delimiters.repetition_separator == "^"
        assert r.delimiters.segment_terminator == "~"

        for _ in r.segments():
            segment_count += 1

        assert 21 == segment_count

    assert r._x12_stream.closed
    assert r.delimiters is None
    assert r._x12_input is None


def test_invalid_x12_data(simple_270_one_line):
    """
    Tests an invalid x12 message input.
    :param simple_270_one_line: X12 Fixture
    """

    invalid_x12 = simple_270_one_line.replace("ISA", "FOO")

    with pytest.raises(ValueError):
        with X12SegmentReader(invalid_x12) as r:
            for _ in r.segments():
                assert False, "Expecting ValueError Exception"


def test_invalid_x12_file(tmpdir, simple_270_one_line):
    """
    Tests an invalid x12 file path input.
    :param tmpdir: Pytest fixture used to create temp files and directories.
    :param simple_270_one_line: X12 Fixture
    """

    invalid_x12 = simple_270_one_line.replace("ISA", "FOO")
    f = tmpdir.mkdir("x12-support").join("test.x12")
    f.write(invalid_x12)

    with pytest.raises(ValueError):
        with X12SegmentReader(f.strpath) as r:
            for _ in r.segments():
                assert False, "Expecting ValueError Exception"


def test_large_x12_message(monkeypatch, large_x12_message, config):
    """
    Tests a large x12 message to validate buffering.
    :param monkeypatch: Pytest fixture for monkeypatching/mocking code.
    :param large_x12_message: Large X12 message fixture
    :param config: X12 configuration fixture
    """

    def mock_get_config():
        return config

    with monkeypatch.context() as m:
        m.setattr(x12.io, "get_config", mock_get_config)

        segment_counter = 0
        with X12SegmentReader(large_x12_message) as r:
            for _ in r.segments():
                segment_counter += 1

        assert r._x12_stream.closed
        assert segment_counter == 85_004



"""
test_common_validations.py

Tests common validations that are shared across X12 transaction sets.
"""
from x12.io import X12ModelReader
import pytest


def test_segment_footer_count(x12_270_subscriber_input):
    """Validates the total segment count validation raises an exception"""
    test_input = x12_270_subscriber_input.replace("SE*27*0001~", "SE*17*0001~")
    with pytest.raises(ValueError):
        with X12ModelReader(test_input) as r:
            for _ in r.models():
                pass



"""
parsing.py

Parses X12 837 005010X222A2 segments into a transactional domain model.

The parsing module includes a specific parser for the 837 transaction and loop parsing functions to create new loops
as segments are streamed to the transactional data model.

Loop parsing functions are implemented as set_[description]_loop(context: X12ParserContext, segment_data: Dict).
"""

from enum import Enum
from x12.parsing import match, X12ParserContext
from typing import Dict, Optional


class TransactionLoops(str, Enum):
    """
    The loops used to support the 837 005010X222A2 format.
    """

    HEADER = "header"
    SUBMITTER_NAME = "loop_1000a"
    RECEIVER_NAME = "loop_1000b"
    BILLING_PROVIDER = "loop_2000a"
    BILLING_PROVIDER_NAME = "loop_2010aa"
    BILLING_PROVIDER_PAY_TO_ADDRESS = "loop_2010ab"
    BILLING_PROVIDER_PAY_TO_PLAN_NAME = "loop_2010ac"
    SUBSCRIBER = "loop_2000b"
    SUBSCRIBER_NAME = "loop_2010ba"
    SUBSCRIBER_PAYER_NAME = "loop_2010bb"
    PATIENT_LOOP = "loop_2000c"
    PATIENT_LOOP_NAME = "loop_2010ca"
    CLAIM_INFORMATION = "loop_2300"
    CLAIM_REFERRING_PROVIDER_NAME = "loop_2310a"
    CLAIM_RENDERING_PROVIDER_NAME = "loop_2310b"
    CLAIM_SERVICE_FACILITY_LOCATION_NAME = "loop_2310c"
    CLAIM_SUPERVISING_PROVIDER_NAME = "loop_2310d"
    CLAIM_AMBULANCE_PICKUP_LOCATION = "loop_2310e"
    CLAIM_AMBULANCE_DROPOFF_LOCATION = "loop_2310f"
    CLAIM_OTHER_SUBSCRIBER_INFORMATION = "loop_2320"
    CLAIM_OTHER_SUBSCRIBER_NAME = "loop_2330a"
    CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_NAME = "loop_2330b"
    CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_REFERRING_PROVIDER_NAME = "loop_2330c"
    CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_RENDERING_PROVIDER_NAME = "loop_2330d"
    CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_SERVICE_FACILITY_LOCATION = "loop_2330e"
    CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_SUPERVISING_PROVIDER_NAME = "loop_2330f"
    CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_BILLING_PROVIDER_NAME = "loop_2330g"
    CLAIM_SERVICE_LINE = "loop_2400"
    CLAIM_SERVICE_LINE_DRUG_IDENTIFICATION_NAME = "loop_2410"
    CLAIM_SERVICE_LINE_RENDERING_PROVIDER_NAME = "loop_2420a"
    CLAIM_SERVICE_LINE_PURCHASED_SERVICE_PROVIDER_NAME = "loop_2420b"
    CLAIM_SERVICE_LINE_SERVICE_FACILITY_LOCATION_NAME = "loop_2420c"
    CLAIM_SERVICE_LINE_SUPERVISING_PROVIDER_NAME = "loop_2420d"
    CLAIM_SERVICE_LINE_ORDERING_PROVIDER_NAME = "loop_2420e"
    CLAIM_SERVICE_LINE_REFERRING_PROVIDER_NAME = "loop_2420f"
    CLAIM_SERVICE_LINE_AMBULANCE_PICKUP_LOCATION = "loop_2420g"
    CLAIM_SERVICE_LINE_AMBULANCE_DROPOFF_LOCATION = "loop_2420h"
    CLAIM_SERVICE_LINE_LINE_AJUDICATION_INFORMATION = "loop_2430"
    CLAIM_SERVICE_LINE_LINE_FORM_IDENTIFICATION = "loop_2440"
    FOOTER = "footer"


def _get_header(context: X12ParserContext) -> Dict:
    """Returns the 837 transaction header"""
    return context.transaction_data[TransactionLoops.HEADER]


def _get_billing_provider(context: X12ParserContext) -> Dict:
    """Returns the current billing provider record"""
    return context.transaction_data[TransactionLoops.BILLING_PROVIDER][-1]


def _get_subscriber(context: X12ParserContext) -> Dict:
    """Returns the current subscriber record"""
    billing_provider = _get_billing_provider(context)
    return billing_provider[TransactionLoops.SUBSCRIBER][-1]


def _get_patient(context: X12ParserContext) -> Optional[Dict]:
    """Returns the current patient record (subscriber or dependent)"""
    subscriber = _get_subscriber(context)
    is_subscriber_patient = (
        subscriber.get("hl_segment", {}).get("hierarchical_child_code", "0") == "0"
    )

    if is_subscriber_patient:
        return subscriber
    else:
        return subscriber.get(TransactionLoops.PATIENT_LOOP, [{}])[-1]


def _is_patient_a_dependent(patient_record: Dict):
    """Returns true if the patient is a dependent record"""
    return patient_record.get("hl_segment", {}).get("hierarchical_level_code") == "23"


def _get_claim(context: X12ParserContext) -> Dict:
    """Returns the current claim record for the patient (either subscriber or dependent)"""
    patient = _get_patient(context)
    if _is_patient_a_dependent(patient):
        claim = patient[TransactionLoops.CLAIM_INFORMATION][-1]
    else:
        subscriber = _get_subscriber(context)
        claim = subscriber[TransactionLoops.CLAIM_INFORMATION][-1]
    return claim


def _get_other_subscriber(context: X12ParserContext) -> Dict:
    """Returns the other subscriber record for the current healthcare claim"""
    claim = _get_claim(context)
    return claim.get(TransactionLoops.CLAIM_OTHER_SUBSCRIBER_INFORMATION, [{}])[-1]


def _get_service_line(context: X12ParserContext) -> Dict:
    """Returns the current claim service line for the patieht (either subscriber or dependent)"""
    claim = _get_claim(context)
    return claim[TransactionLoops.CLAIM_SERVICE_LINE][-1]


@match("ST")
def set_header_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the transaction set header loop for the 837 transaction set.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """

    context.set_loop_context(
        TransactionLoops.HEADER, context.transaction_data[TransactionLoops.HEADER]
    )


@match("NM1", conditions={"entity_identifier_code": "41"})
def set_submitter_name_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the submitter name loop for the 837 transaction set.
    The submitter name loop is nested within the header loop.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if context.loop_name == TransactionLoops.HEADER:
        context.transaction_data[TransactionLoops.SUBMITTER_NAME] = {"per_segment": []}
        submitter_loop = context.transaction_data[TransactionLoops.SUBMITTER_NAME]
        context.set_loop_context(TransactionLoops.SUBMITTER_NAME, submitter_loop)


@match("NM1", conditions={"entity_identifier_code": "40"})
def set_receiver_name_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the receiver name loop for the 837 transaction set.
    The receiver name loop is nested within the header loop.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if context.loop_name == TransactionLoops.SUBMITTER_NAME:
        context.transaction_data[TransactionLoops.RECEIVER_NAME] = {}
        receiver_loop = context.transaction_data[TransactionLoops.RECEIVER_NAME]
        context.set_loop_context(TransactionLoops.RECEIVER_NAME, receiver_loop)


@match("HL", conditions={"hierarchical_level_code": "20"})
def set_billing_provider_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the billing provider loop for the 837 transaction set.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if TransactionLoops.BILLING_PROVIDER not in context.transaction_data:
        context.transaction_data[TransactionLoops.BILLING_PROVIDER] = [{}]
    else:
        context.transaction_data[TransactionLoops.BILLING_PROVIDER].append({})

    billing_provider = context.transaction_data[TransactionLoops.BILLING_PROVIDER][-1]
    context.set_loop_context(TransactionLoops.BILLING_PROVIDER, billing_provider)


@match("NM1", conditions={"entity_identifier_code": "85"})
def set_billing_provider_name_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Sets the billing provider name loop for the 837 transaction set.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if context.loop_name == TransactionLoops.BILLING_PROVIDER:
        billing_provider = _get_billing_provider(context)
        billing_provider[TransactionLoops.BILLING_PROVIDER_NAME] = {"ref_segment": []}
        billing_provider_name = billing_provider[TransactionLoops.BILLING_PROVIDER_NAME]
        context.set_loop_context(
            TransactionLoops.BILLING_PROVIDER_NAME, billing_provider_name
        )


@match("NM1", conditions={"entity_identifier_code": "87"})
def set_pay_to_address_name_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the pay to address name loop for the 837 transaction set.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if context.loop_name == TransactionLoops.BILLING_PROVIDER_NAME:
        billing_provider = _get_billing_provider(context)
        billing_provider[TransactionLoops.BILLING_PROVIDER_PAY_TO_ADDRESS] = {}
        pay_to_address = billing_provider[
            TransactionLoops.BILLING_PROVIDER_PAY_TO_ADDRESS
        ]
        context.set_loop_context(
            TransactionLoops.BILLING_PROVIDER_PAY_TO_ADDRESS, pay_to_address
        )


@match("NM1", conditions={"entity_identifier_code": "PE"})
def set_pay_to_plan_name_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the pay to plan name loop for the 837 transaction set.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if context.loop_name in (
        TransactionLoops.BILLING_PROVIDER_NAME,
        TransactionLoops.BILLING_PROVIDER_PAY_TO_ADDRESS,
    ):
        billing_provider = _get_billing_provider(context)
        billing_provider[TransactionLoops.BILLING_PROVIDER_PAY_TO_PLAN_NAME] = {
            "ref_segment": []
        }
        pay_to_plan = billing_provider[
            TransactionLoops.BILLING_PROVIDER_PAY_TO_PLAN_NAME
        ]
        context.set_loop_context(
            TransactionLoops.BILLING_PROVIDER_PAY_TO_PLAN_NAME, pay_to_plan
        )


@match("HL", conditions={"hierarchical_level_code": "22"})
def set_subscriber_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the subscriber loop for the 837 transaction set.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    billing_provider = _get_billing_provider(context)

    if TransactionLoops.SUBSCRIBER not in billing_provider:
        billing_provider[TransactionLoops.SUBSCRIBER] = [{}]
    else:
        billing_provider[TransactionLoops.SUBSCRIBER].append({})

    subscriber = billing_provider[TransactionLoops.SUBSCRIBER][-1]

    has_dependent_patient = segment_data.get("hierarchical_child_code", "0") == "1"
    if has_dependent_patient and TransactionLoops.PATIENT_LOOP not in subscriber:
        subscriber[TransactionLoops.PATIENT_LOOP] = [{}]

    context.set_loop_context(TransactionLoops.SUBSCRIBER, subscriber)


@match("NM1", conditions={"entity_identifier_code": "IL"})
def set_subscriber_name_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the subscriber name for the 837 transaction set

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if context.loop_name == TransactionLoops.SUBSCRIBER:
        subscriber = _get_subscriber(context)
        subscriber[TransactionLoops.SUBSCRIBER_NAME] = {"ref_segment": []}
        subscriber_name = subscriber[TransactionLoops.SUBSCRIBER_NAME]
        context.set_loop_context(TransactionLoops.SUBSCRIBER_NAME, subscriber_name)


@match("NM1", conditions={"entity_identifier_code": "PR"})
def set_payer_name_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the payer name for the 837 transaction set

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if context.loop_name == TransactionLoops.SUBSCRIBER_NAME:
        subscriber = _get_subscriber(context)
        subscriber[TransactionLoops.SUBSCRIBER_PAYER_NAME] = {"ref_segment": []}
        payer_name = subscriber[TransactionLoops.SUBSCRIBER_PAYER_NAME]
        context.set_loop_context(TransactionLoops.SUBSCRIBER_PAYER_NAME, payer_name)


@match("HL", conditions={"hierarchical_level_code": "23"})
def set_patient_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the patient loop.
    The patient loop is used when the patient is not the subscriber, or cannot be identified as a subscriber.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    subscriber = _get_subscriber(context)
    patient_loop = subscriber[TransactionLoops.PATIENT_LOOP][-1]
    context.set_loop_context(TransactionLoops.PATIENT_LOOP, patient_loop)


@match("NM1", conditions={"entity_identifier_code": "QC"})
def set_patient_name_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the patient name loop

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    patient = _get_patient(context)

    if TransactionLoops.PATIENT_LOOP_NAME not in patient:
        patient[TransactionLoops.PATIENT_LOOP_NAME] = {}

    patient_name = patient[TransactionLoops.PATIENT_LOOP_NAME]
    context.set_loop_context(TransactionLoops.PATIENT_LOOP_NAME, patient_name)


@match("CLM")
def set_claim_information_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the claim information loop for a subscriber or dependent

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    is_dependent_claim = (
        _get_subscriber(context)
        .get("hl_segment", {})
        .get("hierarchical_child_code", "0")
        == "1"
    )

    if not is_dependent_claim:
        patient = _get_subscriber(context)
    else:
        patient = _get_patient(context)

    patient_claim = {
        "dtp_segment": [],
        "ref_segment": [],
        "pwk_segment": [],
        "k3_segment": [],
        "crc_segment": [],
        "hi_segment": [],
    }

    if TransactionLoops.CLAIM_INFORMATION not in patient:
        patient[TransactionLoops.CLAIM_INFORMATION] = [patient_claim]
    else:
        patient[TransactionLoops.CLAIM_INFORMATION].append(patient_claim)

    patient_claim = patient[TransactionLoops.CLAIM_INFORMATION][-1]
    context.set_loop_context(TransactionLoops.CLAIM_INFORMATION, patient_claim)


@match(
    "NM1",
    conditions={"entity_identifier_code": ["DN", "P3", "B2", "77", "DQ", "PW", "45"]},
)
def set_claim_sub_loop(context: X12ParserContext, segment_data: Dict):
    """
    Matches Claim 2300 "sub-loops" 2310A-2310F

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    # confirm we are in loop 2300 or loop 2310x
    if (
        context.loop_name.value.rfind("2300") < 0
        or context.loop_name.value.rfind("2310") < 0
    ):
        return

    identifier_code = segment_data.get("entity_identifier_code")
    claim = _get_claim(context)
    loop_name = None

    if identifier_code in ("PW", "45"):
        loop_data = {}
    else:
        loop_data = {"ref_segment": []}

    if identifier_code in ("DN", "P3"):
        loop_name = TransactionLoops.CLAIM_REFERRING_PROVIDER_NAME
    elif identifier_code == "B2":
        loop_name = TransactionLoops.CLAIM_RENDERING_PROVIDER_NAME
    elif identifier_code == "77":
        loop_name = TransactionLoops.CLAIM_SERVICE_FACILITY_LOCATION_NAME
    elif identifier_code == "DQ":
        loop_name = TransactionLoops.CLAIM_SUPERVISING_PROVIDER_NAME
    elif identifier_code == "PW":
        loop_name = TransactionLoops.CLAIM_AMBULANCE_PICKUP_LOCATION
    elif identifier_code == "45":
        loop_name = TransactionLoops.CLAIM_AMBULANCE_DROPOFF_LOCATION

    if not loop_name:
        raise RuntimeError(f"Unable to parse entity identifier {identifier_code}")

    claim[loop_name] = loop_data
    record = claim[loop_name]
    context.set_loop_context(loop_name, record)


@match("SBR")
def set_other_subscriber(context: X12ParserContext, segment_data: Dict):
    """
    Sets the claim other subscriber loop used to coordinate payment and other insurance carriers.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    # confirm we are in loop 2300 or loop 2310x
    if (
        context.loop_name.value.rfind("2300") > 0
        or context.loop_name.value.rfind("2310") > 0
    ):
        claim = _get_claim(context)
        other_subscriber = {"cas_segment": [], "amt_segment": []}

        if TransactionLoops.CLAIM_OTHER_SUBSCRIBER_INFORMATION not in claim:
            claim[TransactionLoops.CLAIM_OTHER_SUBSCRIBER_INFORMATION] = [
                other_subscriber
            ]
        else:
            claim[TransactionLoops.CLAIM_OTHER_SUBSCRIBER_INFORMATION].append(
                other_subscriber
            )

        other_subscriber = claim[TransactionLoops.CLAIM_OTHER_SUBSCRIBER_INFORMATION][
            -1
        ]
        context.set_loop_context(
            TransactionLoops.CLAIM_OTHER_SUBSCRIBER_INFORMATION, other_subscriber
        )


@match("NM1", conditions={"entity_identifier_code": "IL"})
def set_other_subscriber_name(context: X12ParserContext, segment_data: Dict):
    """
    Sets the claim other subscriber name loop.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    # confirm we are in loop 2300 or loop 2310x
    if (
        context.loop_name.value.rfind("2300") > 0
        or context.loop_name.value.rfind("2310") > 0
    ):
        other_subscriber = _get_other_subscriber(context)

        if TransactionLoops.CLAIM_OTHER_SUBSCRIBER_NAME not in other_subscriber:
            other_subscriber[TransactionLoops.CLAIM_OTHER_SUBSCRIBER_NAME] = {}

        context.set_loop_context(
            TransactionLoops.CLAIM_OTHER_SUBSCRIBER_NAME,
            other_subscriber[TransactionLoops.CLAIM_OTHER_SUBSCRIBER_NAME],
        )


@match(
    "NM1",
    conditions={"entity_identifier_code": ["PR", "DN", "P3", "82", "77", "DQ", "85"]},
)
def set_other_subscriber_other_payer_loop(
    context: X12ParserContext, segment_data: Dict
):
    """
    Sets the other subscriber other payer information for loops 2330B - 2330G

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    # confirm we are in loop 2300 or loop 2310x
    if (
        context.loop_name.value.rfind("2300") < 0
        or context.loop_name.value.rfind("2310") < 0
    ):
        return

    other_subscriber = _get_other_subscriber(context)
    identifier_code = segment_data.get("entity_identifier_code")

    if identifier_code == "PR":
        loop_name = TransactionLoops.CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_NAME
    elif identifier_code in ("DN", "P3"):
        loop_name = (
            TransactionLoops.CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_REFERRING_PROVIDER_NAME
        )
    elif identifier_code == "82":
        loop_name = (
            TransactionLoops.CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_RENDERING_PROVIDER_NAME
        )
    elif identifier_code == "77":
        loop_name = (
            TransactionLoops.CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_SERVICE_FACILITY_LOCATION
        )
    elif identifier_code == "DQ":
        loop_name = (
            TransactionLoops.CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_SUPERVISING_PROVIDER_NAME
        )
    elif identifier_code == "85":
        loop_name = (
            TransactionLoops.CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_BILLING_PROVIDER_NAME
        )
    else:
        loop_name = None

    if not loop_name:
        raise RuntimeError(
            f"Unable to assign loop to other subscriber identifier {identifier_code}"
        )

    if (
        loop_name
        == TransactionLoops.CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_REFERRING_PROVIDER_NAME
    ):
        if loop_name not in other_subscriber:
            other_subscriber[loop_name] = [{"ref_segment": []}]
        else:
            other_subscriber[loop_name].append({"ref_segment": {}})
        loop_record = other_subscriber[loop_name][-1]
    else:
        other_subscriber[loop_name] = {"ref_segment": []}
        loop_record = other_subscriber[loop_name]

    context.set_loop_context(loop_name, loop_record)


@match("LX")
def set_service_line_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the claim service line loop
    """
    claim = _get_claim(context)
    new_service_line = {
        "pwk_segment": [],
        "crc_segment": [],
        "dtp_segment": [],
        "qty_segment": [],
        "mea_segment": [],
        "ref_segment": [],
        "amt_segment": [],
        "k3_segment": [],
        "nte_segment": [],
    }
    if TransactionLoops.CLAIM_SERVICE_LINE not in claim:
        claim[TransactionLoops.CLAIM_SERVICE_LINE] = [new_service_line]
    else:
        claim[TransactionLoops.CLAIM_SERVICE_LINE].append(new_service_line)

    service_line = claim[TransactionLoops.CLAIM_SERVICE_LINE][-1]
    context.set_loop_context(TransactionLoops.CLAIM_SERVICE_LINE, service_line)


@match("LIN")
def set_drug_identification_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the Claim Service Line (Loop 2400) Drug Identification Loop

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    claim = _get_claim(context)

    if TransactionLoops.CLAIM_SERVICE_LINE not in claim:
        claim[TransactionLoops.CLAIM_SERVICE_LINE] = [{}]
    else:
        claim[TransactionLoops.CLAIM_SERVICE_LINE].append([{}])

    service_line = claim[TransactionLoops.CLAIM_SERVICE_LINE][-1]
    context.set_loop_context(TransactionLoops.CLAIM_SERVICE_LINE, service_line)


@match(
    "NM1",
    conditions={"entity_identifier_code": ["82", "QB", "77", "DQ", "DK", "DN", "P3"]},
)
def set_service_line_entity_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets entity loops (rendering provider, order provider, etc) within Loop 2400/Service Line
    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if not context.loop_name.value.rfind("loop_24"):
        return

    service_line = _get_service_line(context)
    if service_line:
        loop_name = None
        loop_data = {"ref_segment": []}
        identifier_code = segment_data.get("entity_identifier_code")

        if identifier_code == "82":
            loop_name = TransactionLoops.CLAIM_SERVICE_LINE_RENDERING_PROVIDER_NAME
        elif identifier_code == "QB":
            loop_name = (
                TransactionLoops.CLAIM_SERVICE_LINE_PURCHASED_SERVICE_PROVIDER_NAME
            )
        elif identifier_code == "77":
            loop_name = (
                TransactionLoops.CLAIM_SERVICE_LINE_SERVICE_FACILITY_LOCATION_NAME
            )
        elif identifier_code == "DQ":
            loop_name = TransactionLoops.CLAIM_SERVICE_LINE_SUPERVISING_PROVIDER_NAME
        elif identifier_code == "DK":
            loop_name = TransactionLoops.CLAIM_SERVICE_LINE_ORDERING_PROVIDER_NAME
        elif identifier_code in ("DN", "P3"):
            loop_name = TransactionLoops.CLAIM_SERVICE_LINE_REFERRING_PROVIDER_NAME

        if loop_name is None:
            raise RuntimeError(f"unknown identifier {identifier_code}")

        context.set_loop_context(loop_name, loop_data)


@match("SVD")
def set_line_adjudication_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the claim service line adjudication loop

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    service_line = _get_service_line(context)

    loop_data = {"cas_segment": []}
    loop_name = TransactionLoops.CLAIM_SERVICE_LINE_LINE_AJUDICATION_INFORMATION

    if loop_name not in service_line:
        service_line[loop_name] = loop_data
    else:
        service_line[loop_name].append(loop_data)

    loop_record = service_line[loop_name][-1]

    context.set_loop_context(loop_name, loop_record)


@match("LQ")
def set_form_identification_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the form identification loop.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    service_line = _get_service_line(context)
    loop_name = TransactionLoops.CLAIM_SERVICE_LINE_LINE_FORM_IDENTIFICATION
    loop_data = {"frm_segment": []}

    if loop_name not in service_line:
        service_line[loop_name] = [loop_data]
    else:
        service_line[loop_name].append(loop_data)

    loop_record = service_line[loop_name][-1]
    context.set_loop_context(loop_name, loop_record)


@match("SE")
def set_se_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the transaction set footer loop.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """

    context.set_loop_context(
        TransactionLoops.FOOTER, context.transaction_data["footer"]
    )



"""
segments.py

Specialized segment models for the HealthCare Claim Professional 837 005010X222A2 transaction.
"""

from x12.segments import (
    StSegment,
    BhtSegment,
    Nm1Segment,
    PerSegment,
    HlSegment,
    PrvSegment,
    RefSegment,
    SbrSegment,
    DtpSegment,
    Cn1Segment,
    AmtSegment,
    NteSegment,
    CrcSegment,
    PatSegment,
    HcpSegment,
    QtySegment,
    CasSegment,
)
from typing import Literal, Optional, Dict
from enum import Enum
from pydantic import Field, root_validator


class HeaderStSegment(StSegment):
    """
    Customized ST segment for 837 transaction header
    """

    transaction_set_identifier_code: Literal["837"]
    implementation_convention_reference: Literal["005010X222A2"]


class HeaderBhtSegment(BhtSegment):
    """
    Customized BHT segment for 837 transaction header
    """

    class PurposeCode(str, Enum):
        """
        BHT02 code values
        """

        ORIGINAL = "00"
        REISSUE = "18"

    class TransactionTypeCode(str, Enum):
        """
        BHT06 code values
        """

        SUBROGATION_DEMAND = "31"
        CHARGABLE = "CH"
        REPORTING = "RP"

    hierarchical_structure_code: Literal["0019"]
    transaction_set_purpose_code: PurposeCode
    transaction_type_code: TransactionTypeCode


class Loop1000ANm1Segment(Nm1Segment):
    """
    Submitter Name Name Segment
    """

    entity_identifier_code: Literal["41"]
    identification_code_qualifier: Literal["46"]


class Loop1000APerSegment(PerSegment):
    """
    Submitter Name Contact Segment
    """

    class CommunicationNumberQualifier1(str, Enum):
        """
        Code values for PER03
        """

        ELECTRONIC_MAIL = "EM"
        FACSIMILE = "FX"
        TELEPHONE = "TE"

    class CommunicationNumberQualifier2(str, Enum):
        """
        Code values for PER05
        """

        ELECTRONIC_MAIL = "EM"
        TELEPHONE_EXTENSION = "EX"
        FACSIMILE = "FX"
        TELEPHONE = "TE"

    communication_number_qualifier_1: CommunicationNumberQualifier1
    communication_number_qualifier_2: Optional[CommunicationNumberQualifier2]
    communication_number_qualifier_3: Optional[CommunicationNumberQualifier2]


class Loop1000BNm1Segment(Nm1Segment):
    """
    Information Receiver Name
    """

    entity_identifier_code: Literal["40"]
    identification_code_qualifier: Literal["46"]


class Loop2000AHlSegment(HlSegment):
    """
    Billing Provider Hierarchical Level
    """

    hierarchical_parent_id_number: Optional[str]
    hierarchical_level_code: Literal["20"]


class Loop2000APrvSegment(PrvSegment):
    """
    Billing Provider Speciality Information
    """

    provider_code: Literal["BI"]
    reference_identification_qualifier: Literal["PXC"]
    reference_identification: str = Field(min_length=1, max_length=50)


class Loop2000BHlSegment(HlSegment):
    """
    Subscriber Hierarchy Segment
    """

    hierarchical_level_code: Literal["22"]


class Loop2000BSbrSegment(SbrSegment):
    """
    Subscriber Information Segment

    TODO: validation rules
    """

    class PayerResponsibilityCode(str, Enum):
        """
        SBR01 code values
        """

        PAYER_RESPONSIBILITY_FOUR = "A"
        PAYER_RESPONSIBILITY_FIVE = "B"
        PAYER_RESPONSIBILITY_SIX = "C"
        PAYER_RESPONSIBILITY_SEVEN = "D"
        PAYER_RESPONSIBILITY_EIGHT = "E"
        PAYER_RESPONSIBILITY_NINE = "F"
        PAYER_RESPONSIBILITY_TEN = "G"
        PRIMARY = "P"
        SECONDARY = "S"
        TERTIARY = "T"
        UNKNOWN = "U"

    class InsuranceTypeCode(str, Enum):
        """
        SBR05 code values
        """

        MEDICARE_SECONDARY_WORKING = "12"
        MEDICARE_SECONDARY_END_STAGE_RENAL = "13"
        MEDICARE_SECONDARY_AUTO_PRIMARY = "14"
        MEDICARE_SECONDARY_WORKERS_COMPENSATION = "15"
        MEDICARE_SECONDARY_PUBLIC_HEALTH_SERVICE = "16"
        MEDICARE_SECONDARY_BLACK_LUNG = "41"
        MEDICARE_SECONDARY_VETERANS_ADMINISTRATION = "42"
        MEDICARE_SECONDARY_DISABLED_BENEFICIARY = "43"
        MEDICARE_SECONDARY_LIABILITY_PRIMARY = "47"

    class ClaimFilingIndicatorCode(str, Enum):
        """
        SBR09 code values
        """

        OTHER_NON_FEDERAL_PROGRAMS = "11"
        PREFERRED_PROVIDER_ORGANIZATION = "12"
        POINT_OF_SERVICE = "13"
        EXCLUSIVE_PROVIDER_ORGANIZATION = "14"
        INDEMNITY_INSURANCE = "15"
        HEALTH_MAINTENANCE_ORGANIZATION_MEDICARE_RISK = "16"
        DENTAL_MAINTENANCE_ORGANIZATION = "17"
        AUTOMOBILE_MEDICAL = "AM"
        BLUE_CROSS_BLUE_SHIELD = "BL"
        CHAMPUS = "CH"
        COMMERICIAL_INSURANCE_COMPANY = "CI"
        DISABILITY = "DS"
        FEDERAL_EMPLOYEES_PROGRAM = "FI"
        HEALTH_MAINTENANCE_ORGANIZATION = "HM"
        LIABILITY_MEDICAL = "LM"
        MEDICARE_PART_A = "MA"
        MEDICARE_PART_B = "MB"
        MEDICAID = "MC"
        OTHER_FEDERAL_PROGRAM = "OF"
        TITLE_V = "TV"
        VETERANS_AFFAIR_PLAN = "VA"
        WORKERS_COMPENSATION_HEALTH_CLAIM = "WC"
        MUTUALLY_DEFINED = "ZZ"

    payer_responsibility_code: PayerResponsibilityCode
    individual_relationship_code: Optional[Literal["18"]]
    insurance_type_code: Optional[InsuranceTypeCode]


class Loop2000CHlSegment(HlSegment):
    """
    The HL segment for Loop 2000C (Patient)
    """

    hierarchical_level_code: Literal["23"]


class Loop2000CPatSegment(PatSegment):
    """
    The PAT, patient, segment for Loop2000C
    """

    class IndividualRelationshipCode(str, Enum):
        """
        Code values for PAT01
        """

        SPOUSE = "01"
        CHILD = "19"
        EMPLOYEE = "20"
        UNKNOWN = "21"
        ORGAN_DONOR = "39"
        CADAVER_DONOR = "40"
        LIFE_PARTNER = "53"
        OTHER_RELATIONSHIP = "G8"

    individual_relationship_code: IndividualRelationshipCode


class Loop2010AaNm1Segment(Nm1Segment):
    """
    Billing Provider Name and Identification
    """

    entity_identifier_code: Literal["85"]
    identification_code_qualifier: Optional[Literal["XX"]]


class Loop2010AaRefSegment(RefSegment):
    """
    Billing Provider Identification Codes
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        EMPLOYER_IDENTIFICATION_NUMBER = "EI"
        SOCIAL_SECURITY_NUMBER = "SY"
        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2010AbNm1Segment(Nm1Segment):
    """
    Billing Provider Name and Identification
    """

    entity_identifier_code: Literal["87"]
    name_last_or_organization_name: Optional[str]


class Loop2010AcNm1Segment(Nm1Segment):
    """
    Pay to Plan Name and Identification
    """

    class IdentificationCodeQualifier(str, Enum):
        """
        Code values for NM108
        """

        PAYOR_IDENTIFICATION = "PI"
        CMS_PLAN_ID = "XV"

    entity_identifier_code: Literal["PE"]
    entity_type_qualifier: Literal["2"]
    name_last_or_organization_name: Optional[str]
    identification_code_qualifier: IdentificationCodeQualifier


class Loop2010AcRefSegment(RefSegment):
    """
    Pay to Plan Identification Codes
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        PAYER_IDENTIFICATION_NUMBER = "2u"
        CLAIM_OFFICE_NUMBER = "FY"
        NAIC_CODE = "NF"
        EMPLOYERS_IDENTIFICATION_NUMBER = "EI"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2010BaNm1Segment(Nm1Segment):
    """
    Subscriber Name Segment
    """

    class IdentificationCodeQualifier(str, Enum):
        """
        Code values for NM108
        """

        STANDARD_HEALTH_ID_US = "II"
        MEMBER_IDENTIFICATION_NUMBER = "MI"

    entity_identifier_code: Literal["IL"]
    identification_code_qualifier: IdentificationCodeQualifier


class Loop2010BaRefSegment(RefSegment):
    """
    Subscriber Name Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        SOCIAL_SECURITY_NUMBER = "SY"
        AGENCY_CLAIM_NUMBER = "Y4"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2010BaPerSegment(PerSegment):
    """
    Subscriber Name Contact Information
    """

    communication_number_qualifier_1: Literal["TE"]
    communication_number_qualifier_2: Optional[Literal["EX"]]


class Loop2010BbNm1Segment(Nm1Segment):
    """
    Subscriber Name Segment
    """

    class IdentificationCodeQualifier(str, Enum):
        """
        Code values for NM108
        """

        PAYOR_IDENTIFICATION = "PI"
        CMS_PLAN_ID = "XV"

    entity_identifier_code: Literal["PR"]
    entity_type_qualifier: Literal["2"]
    identification_code_qualifier: IdentificationCodeQualifier


class Loop2010BbRefSegment(RefSegment):
    """
    Payer Name Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        PAYER_IDENTIFICATION_NUMBER = "2U"
        EMPLOYERS_IDENTIFICATION_NUMBER = "EI"
        CLAIM_OFFICER_NUMBER = "FY"
        NAIC_CODE = "NF"
        PROVIDER_CONTROL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2010CaNm1Segment(Nm1Segment):
    """
    Loop2010CA NM1 segment for Patient
    """

    entity_identifier_code: Literal["QC"]
    entity_type_qualifier: Literal["1"]


class Loop2010CaPerSegment(PerSegment):
    """
    Loop2010CA (Patient Name) Per (Contact) segment
    """

    communication_number_qualifier_1: Literal["TE"]
    communication_number_qualifier_2: Optional[Literal["EX"]]


class Loop2010CaRefSegment(RefSegment):
    """
    Loop 2010CA (Patient Name) REF Segment
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        AGENCY_CLAIM_NUMBER = "Y4"
        MEMBER_IDENTIFICATION_NUMBER = "1W"
        SOCIAL_SECURITY_NUMBER = "SY"


class Loop2300DtpSegment(DtpSegment):
    """
    Claim information DTP segments
    """

    class DateTimeQualifier(str, Enum):
        """
        Code values for DTP01
        """

        ONSET_OF_CURRENT_SYMPTOM_OR_ILLNESS = "431"
        INITIAL_TREATMENT = "454"
        LAST_VISIT_OR_CONSULTATION = "304"
        ACUTE_MANIFESTATION_OF_CHRONIC_CONDITION = "453"
        ACCIDENT = "439"
        LAST_MENSTRUAL_PERIOD = "484"
        LAST_XRAY = "455"
        PRESCRIPTION = "471"
        DISABILITY = "314"
        INITIAL_DISABILITY_START = "360"
        INITIAL_DISABILITY_END = "361"
        INITIAL_DISABILITY_LAST_DATE_WORKED = "297"
        INITIAL_DISABILITY_RETURN_TO_WORK = "296"
        ADMISSION = "435"
        DISCHARGE = "096"
        ASSUMED_CARE_DATE = "090"
        RELINQUISHED_CARE_DATE = "091"
        FIRST_VISIT_CONSULTATION = "444"
        REPRICER_RECEIVED_DATE = "050"

    class DateTimePeriodFormatQualifier(str, Enum):
        """
        Code values for DTP02
        """

        DATE = "D8"
        DATE_RANGE = "RD8"

    date_time_qualifier: DateTimeQualifier
    date_time_period_format_qualifier: DateTimePeriodFormatQualifier

    @root_validator
    def validate_disability_dates(cls, values: Dict):
        """
        Validates that a date range qualifier is used for disability dates.

        :param values: The model's values
        :return: The model's values
        """
        date_qualifier = values.get("date_time_qualifier")
        period_qualifier = values.get("date_time_period_format_qualifier")

        if date_qualifier == "314" and period_qualifier != "RD8":
            raise ValueError(
                "RD8 Date Time Period is required for Disability Dates (314)"
            )
        return values


class Loop2300Cn1Segment(Cn1Segment):
    """
    Claim information contract information
    """

    class ContractTypeCode(str, Enum):
        """
        Code values for CN101
        """

        DIAGNOSIS_RELATED_GROUP = "01"
        PER_DIEM = "02"
        VARIABLE_PER_DIEM = "03"
        FIAT = "04"
        CAPITATED = "05"
        PERCENT = "06"
        OTHER = "09"

    contract_type_code: ContractTypeCode


class Loop2300AmtSegment(AmtSegment):
    """
    Monetary Amount
    """

    amount_qualifier_code: Literal["F5"]


class Loop2300RefSegment(RefSegment):
    """
    Claim information reference identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        SPECIAL_PAYMENT_REFERENCE_NUMBER = "4N"
        MEDICARE_VERSION_CODE = "F5"
        MAMMOGRAPHY_CERTIFICATION_NUMBER = "EW"
        REFERRAL_NUMBER = "9F"
        PRIOR_AUTHORIZATION_NUMBER = "G1"
        ORIGINAL_REFERENCE_NUMBER = "F8"
        CLINICAL_LABORATORY_IMPROVEMENT_AMENDMENT_NUMBER = "X4"
        REPRICED_CLAIM_REFERENCE_NUMBER = "9A"
        ADJUSTED_REPRICED_CLAIM_REFERENCE_NUMBER = "9C"
        QUALIFIED_PRODUCTS_LIST = "LX"
        CLAIM_NUMBER = "D9"
        MEDICAL_RECORD_IDENTIFICATION_NUMBER = "EA"
        PROJECT_CODE = "P4"
        FACILITY_ID_NUMBER = "1J"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2300NteSegment(NteSegment):
    """
    Claim information note segment
    """

    class NoteReferenceCode(str, Enum):
        """
        Code values for NTE01
        """

        ADDITIONAL_INFORMATION = "ADD"
        CERTIFICATION_NARRATIVE = "CER"
        GOALS_REHAB_POTENTIAL_DISCHARGE_PLANS = "DCP"
        DIAGNOSIS_DESCRIPTION = "DGN"
        THIRD_PARTY_ORGANIZATION_NOTES = "TPO"


class Loop2300CrcAmbulanceCertification(CrcSegment):
    """
    Claim Information - Ambulance Certification
    """

    class YesNoResponseCode(str, Enum):
        """
        Code values for CRC02
        """

        NO = "N"
        YES = "Y"

    class ConditionsIndicator(str, Enum):
        """
        Code values for CRC03 - CRC07
        """

        PATIENT_ADMITTED_TO_HOSPITAL = "01"
        PATIENT_MOVED_BY_STRETCHER = "04"
        PATIENT_UNCONSCIOUS_IN_SHOCK = "05"
        PATIENT_EMERGENCY_TRANSPORT = "06"
        PATIENT_RESTRAINED = "07"
        PATIENT_HEMORRHAGING = "08"
        AMBULANCE_MEDICALLY_NECESSARY = "09"
        PATIENT_CONFINED_BED_CHAIR = "12"

    code_category: Literal["07"]
    certification_condition_indicator: YesNoResponseCode
    condition_code_1: ConditionsIndicator
    condition_code_2: Optional[ConditionsIndicator]
    condition_code_3: Optional[ConditionsIndicator]
    condition_code_4: Optional[ConditionsIndicator]
    condition_code_5: Optional[ConditionsIndicator]


class Loop2300CrcPatientConditionVision(CrcSegment):
    """
    Claim Information - Patient Vision Condition
    """

    class CodeCategoryValues(str, Enum):
        """
        Code values for CRC01
        """

        SPECTACLE_LENSES = "E1"
        CONTACT_LENSES = "E2"
        SPECTACLE_FRAMES = "E3"

    class YesNoResponseCode(str, Enum):
        """
        Code values for CRC02
        """

        NO = "N"
        YES = "Y"

    class ConditionsIndicator(str, Enum):
        """
        Code values for CRC03 - CRC07
        """

        GENERAL_STANDARD_20_DEGREE = "L1"
        REPLACEMENT_LOSS_THEFT = "L2"
        REPLACEMENT_BREAKAGE_DAMAGE = "L3"
        REPLACEMENT_PATIENT_PREFERENCE = "L4"
        REPLACEMENT_MEDICAL_REASON = "L5"

    code_category: CodeCategoryValues
    certification_condition_indicator: YesNoResponseCode
    condition_code_1: ConditionsIndicator
    condition_code_2: Optional[ConditionsIndicator]
    condition_code_3: Optional[ConditionsIndicator]
    condition_code_4: Optional[ConditionsIndicator]
    condition_code_5: Optional[ConditionsIndicator]


class Loop2300CrcHomeboundIndicator(CrcSegment):
    """
    Claim Information - Homebound indicator
    """

    class ConditionsIndicator(str, Enum):
        """
        Code values for CRC03 - CRC07
        """

        INDEPENDENT_AT_HOME = "IH"

    code_category: Literal["75"]
    certification_condition_indicator: Literal["Y"]
    condition_code_1: ConditionsIndicator
    condition_code_2: Optional[ConditionsIndicator]
    condition_code_3: Optional[ConditionsIndicator]
    condition_code_4: Optional[ConditionsIndicator]
    condition_code_5: Optional[ConditionsIndicator]


class Loop2300CrcEpSdtRefferal(CrcSegment):
    """
    Claim Information - Homebound indicator
    """

    class YesNoResponseCode(str, Enum):
        """
        Code values for CRC02
        """

        NO = "N"
        YES = "Y"

    class ConditionsIndicator(str, Enum):
        """
        Code values for CRC03 - CRC07
        """

        AVAILABLE_NOT_USED = "AV"
        NOT_USED = "NU"
        UNDER_TREATMENT = "S2"
        NEW_SERVICES_REQUESTED = "ST"

    code_category: Literal["ZZ"]
    certification_condition_indicator: YesNoResponseCode
    condition_code_1: ConditionsIndicator
    condition_code_2: Optional[ConditionsIndicator]
    condition_code_3: Optional[ConditionsIndicator]
    condition_code_4: Optional[ConditionsIndicator]
    condition_code_5: Optional[ConditionsIndicator]


class Loop2300CrcSegment(CrcSegment):
    """
    Claim information conditions indicators
    """

    @root_validator
    def validate_specialized_crc_segment(cls, values):
        """
        Parses the CRC segments code category to determine which model is used for validation.
        """
        code_category = values.get("code_category")

        if code_category == "07":
            Loop2300CrcAmbulanceCertification(**values)
        elif code_category in ("E1", "E2", "E3"):
            Loop2300CrcPatientConditionVision(**values)
        elif code_category == "75":
            Loop2300CrcHomeboundIndicator(**values)
        elif code_category == "ZZ":
            Loop2300CrcEpSdtRefferal(**values)
        else:
            raise ValueError(f"Unknown CRC01 code category value {code_category}")
        return values


class Loop2310ANm1Segment(Nm1Segment):
    """
    Claim Referring Provider Name
    """

    class EntityIdentifierCode(str, Enum):
        """
        Code values for NM101
        """

        REFERRING_PROVIDER = "DN"
        PRIMARY_CARE_PROVIDER = "P3"

    entity_identifier_code: EntityIdentifierCode
    entity_type_qualifier: Literal["1"]
    identification_code_qualifier: Optional[Literal["XX"]]


class Loop2310ARefSegment(RefSegment):
    """
    Claim Referring Provider Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2310BNm1Segment(Nm1Segment):
    """
    Claim Rendering Provider Name
    """

    entity_identifier_code: Literal["82"]
    identification_code_qualifier: Literal["XX"]


class Loop2310BPrvSegment(PrvSegment):
    """
    Claim Rendering Provider Specialty
    """

    provider_code: Literal["PE"]
    reference_identification_qualifier: Literal["PXC"]
    reference_identification: str = Field(min_length=1, max_length=50)


class Loop2310BRefSegment(RefSegment):
    """
    Claim Rendering Provider Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2310CaPerSegment(PerSegment):
    """
    Loop2310CA (Patient Name) Per (Contact) segment
    """

    communication_number_qualifier_1: Literal["TE"]
    communication_number_qualifier_2: Optional[Literal["EX"]]


class Loop2310CNm1Segment(Nm1Segment):
    """
    Claim Service Facility Location Name
    """

    entity_identifier_code: Literal["77"]
    entity_type_qualifier: Literal["2"]
    identification_code_qualifier: Literal["XX"]


class Loop2310CRefSegment(RefSegment):
    """
    Claim Service Facility Location Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2310CPerSegment(PerSegment):
    """
    Claim Service Facility Contact Segment
    """

    communication_number_qualifier_1: Literal["TE"]
    communication_number_qualifier_2: Optional[Literal["EX"]]


class Loop2310DNm1Segment(Nm1Segment):
    """
    Claim Supervising Provider Name
    """

    entity_identifier_code: Literal["DQ"]
    entity_type_qualifier: Literal["1"]
    identification_code_qualifier: Literal["XX"]


class Loop2310DRefSegment(Nm1Segment):
    """
    Claim Supervising Provider Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2310ENm1Segment(Nm1Segment):
    """
    Claim ambulance pickup location
    """

    entity_identifier_code: Literal["PW"]
    entity_type_qualifier: Literal["2"]


class Loop2310FNm1Segment(Nm1Segment):
    """
    Claim ambulance pickup location
    """

    entity_identifier_code: Literal["45"]
    entity_type_qualifier: Literal["2"]


class Loop2320SbrSegment(SbrSegment):
    """
    Claim Other Subscriber
    """

    class PayerResponsibilityCode(str, Enum):
        """
        Code values for SBR01
        """

        PAYER_RESPONSIBILITY_FOUR = "A"
        PAYER_RESPONSIBILITY_FIVE = "B"
        PAYER_RESPONSIBILITY_SIX = "C"
        PAYER_RESPONSIBILITY_SEVEN = "D"
        PAYER_RESPONSIBILITY_EIGHT = "E"
        PAYER_RESPONSIBILITY_NINE = "F"
        PAYER_RESPONSIBILITY_TEN = "G"
        PAYER_RESPONSIBILITY_ELEVEN = "H"
        PAYER_RESPONSIBILITY_PRIMARY = "P"
        PAYER_RESPONSIBILITY_SECONDARY = "S"
        PAYER_RESPONSIBILITY_TERTIARY = "T"
        PAYER_RESPONSIBILITY_UNKNOWN = "U"

    class IndividualRelationshipCode(str, Enum):
        """
        Code values for SBR02
        """

        SPOUSE = "01"
        SELF = "18"
        CHILD = "19"
        EMPLOYEE = "20"
        UNKNOWN = "21"
        ORGAN_DONOR = "39"
        CADAVER_DONOR = "40"
        LIFE_PARTNER = "53"
        OTHER_RELATIONSHIP = "G8"

    class InsuranceTypeCode(str, Enum):
        """
        SBR05 code values
        """

        MEDICARE_SECONDARY_WORKING = "12"
        MEDICARE_SECONDARY_END_STAGE_RENAL = "13"
        MEDICARE_SECONDARY_AUTO_PRIMARY = "14"
        MEDICARE_SECONDARY_WORKERS_COMPENSATION = "15"
        MEDICARE_SECONDARY_PUBLIC_HEALTH_SERVICE = "16"
        MEDICARE_SECONDARY_BLACK_LUNG = "41"
        MEDICARE_SECONDARY_VETERANS_ADMINISTRATION = "42"
        MEDICARE_SECONDARY_DISABLED_BENEFICIARY = "43"
        MEDICARE_SECONDARY_LIABILITY_PRIMARY = "47"


class Loop2330aNm1Segment(Nm1Segment):
    """
    Claim Other Subscriber Name
    """

    class IdentificationCodeQualifier(str, Enum):
        """
        Code values for NM108
        """

        STANDARD_HEALTH_ID_US = "II"
        MEMBER_IDENTIFICATION_NUMBER = "MI"

    entity_identifier_code: Literal["IL"]


class Loop2330aRefSegment(RefSegment):
    """
    Claim Other Subscriber Reference Identification
    """

    reference_identification_qualifier: Literal["SY"]


class Loop2330bNm1Segment(Nm1Segment):
    """
    Claim Other Subscriber Other Payer Name
    """

    class IdentificationCodeQualifier(str, Enum):
        """
        Code values for NM108
        """

        PAYOR_IDENTIFICATION = "PI"
        CMS_PLAN_ID = "XV"

    entity_identifier_code: Literal["PR"]
    entity_type_qualifier: Literal["2"]
    name_last_or_organization_name: str = Field(min_length=1, max_length=60)
    identification_code_qualifier: IdentificationCodeQualifier
    identification_code: str = Field(min_length=2, max_length=80)


class Loop2330BDtpSegment(DtpSegment):
    """
    Claim Other Subscriber Other Payer Claim Check or Remittance Date
    """

    date_time_qualifier: Literal["573"]
    date_time_period_format_qualifier: Literal["D8"]


class Loop2300BRefSegment(RefSegment):
    """
    Claim Other Subscriber Other Payer Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        PAYOR_IDENTIFICATION = "PI"
        EMPLOYER_IDENTIFICATION_NUMBER = "EI"
        CLAIM_OFFICE_NUMBER = "FY"
        NAIC_CODE = "NF"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2330cNm1Segment(Nm1Segment):
    """
    Claim Other Subscriber Other Payer Referring Provider Name
    """

    class EntityIdentifierCode(str, Enum):
        """
        Code values for NM01
        """

        REFERRING_PROVIDER = "DN"
        PRIMARY_CARE_PROVIDER = "P3"

    entity_identifier_code: EntityIdentifierCode
    entity_type_qualifier: Literal["1"]


class Loop2330cRefSegment(RefSegment):
    """
    Claim Other Subscriber Referring Provider
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2330dNm1Segment(Nm1Segment):
    """
    Claim Other Subscriber Other Payer Rendering Provider Name
    """

    entity_identifier_code: Literal["82"]
    name_last_or_organization_name: Optional[str]


class Loop2330dRefSegment(RefSegment):
    """
    Claim Other Subscriber Rendering Provider
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2330eNm1Segment(Nm1Segment):
    """
    Claim Other Subscriber Other Payer Service Facility Location
    """

    entity_identifier_code: Literal["77"]
    entity_type_qualifier: Literal["2"]


class Loop2330eRefSegment(RefSegment):
    """
    Claim Other Subscriber Other Payer Service Facility Location Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2330fNm1Segment(Nm1Segment):
    """
    Claim Other Subscriber Other Payer Service Facility Location
    """

    entity_identifier_code: Literal["DQ"]
    entity_type_qualifier: Literal["1"]


class Loop2330fRefSegment(RefSegment):
    """
    Claim Other Subscriber Other Payer Service Facility Location Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2330gNm1Segment(Nm1Segment):
    """
    Claim Other Subscriber Other Payer Service Facility Location
    """

    entity_identifier_code: Literal["85"]


class Loop2330gRefSegment(RefSegment):
    """
    Claim Other Subscriber Other Payer Service Facility Location Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2400CrcSegment(CrcSegment):
    """
    Conditions Indicator
    """

    class YesNoResponseCode(str, Enum):
        """
        Code values for CRC
        """

        YES = "Y"
        NO = "N"

    class CodeCategory(str, Enum):
        """
        Code values for CRC01
        """

        DURABLE_MEDICAL_EQUIPMENT_CERTIFICATION = "09"
        HOSPICE = "70"

    class ConditionsIndicator(str, Enum):
        """
        Code values for CRC03 - CRC07
        """

        # Ambulance Conditions
        PATIENT_ADMITTED_TO_HOSPITAL = "01"
        PATIENT_MOVED_BY_STRETCHER = "04"
        PATIENT_UNCONSCIOUS_IN_SHOCK = "05"
        PATIENT_EMERGENCY_TRANSPORT = "06"
        PATIENT_RESTRAINED = "07"
        PATIENT_HEMORRHAGING = "08"
        AMBULANCE_MEDICALLY_NECESSARY = "09"
        PATIENT_CONFINED_BED_CHAIR = "12"
        # Hospice Condition
        OPEN = "65"
        # Durable Medical Equipment Condition
        CERTIFICATE_ON_FILE_SUPPLIER = "38"
        REPLACEMENT_ITEM = "ZV"

    code_category: CodeCategory
    certification_condition_indicator: YesNoResponseCode
    condition_code_1: ConditionsIndicator
    condition_code_2: Optional[ConditionsIndicator]
    condition_code_3: Optional[ConditionsIndicator]
    condition_code_4: Optional[ConditionsIndicator]
    condition_code_5: Optional[ConditionsIndicator]


class Loop2400DtpSegment(DtpSegment):
    """
    Date values for Loop 2400 - Claim service line
    """

    class DateTimeQualifier(str, Enum):
        """
        Code values for DTP001
        """

        SERVICE = "472"
        PRESCRIPTION = "471"
        CERTIFICATION_REVISION = "607"
        BEGIN_THERAPY = "463"
        LAST_CERTIFICATION = "461"
        LATEST_VISIT_CONSULTATION = "304"
        MOST_RECENT_HEMOGLOBIN_HEMATOCRIT = "738"
        MOST_RECENT_SERUM_CREATINE = "739"
        SHIPPED = "011"
        LAST_X_RAY = "455"
        INITIAL_TREATMENT = "454"

    date_time_qualifier: DateTimeQualifier


class Loop2400QtySegment(QtySegment):
    """
    Quantity Information for Claim Service Lines
    """

    class QuantityQualifier(str, Enum):
        """
        Code values for QTY01
        """

        UNITS = "FL"
        PATIENTS = "PL"

    quantity_qualifier: QuantityQualifier


class Loop2400Cn1Segment(Cn1Segment):
    """
    Contract Information
    """

    class ContractTypeCode(str, Enum):
        """
        Code values for CN101
        """

        DIAGNOSIS_RELATED_GROUP = "01"
        PER_DIEM = "02"
        VARIABLE_PER_DIEM = "03"
        FLAT = "04"
        CAPITATED = "05"
        PERCENT = "06"
        OTHER = "09"


class Loop2400RefSegment(RefSegment):
    """
    Loop 2400/Service Line Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        REPRICED_LINE_ITEM_REFERENCE_NUMBER = "9B"
        ADJUSTED_REPRICED_LINE_ITEM_REFERENCE_NUMBER = "9D"
        PRIOR_AUTHORIZATION_NUMBER = "G1"
        PROVIDER_CONTROL_NUMBER = "6R"
        MAMMOGRAPHY_CERTIFICATION_NUMBER = "EW"
        CLINICAL_LABORATORY_IMPROVEMENT_NUMBER = "X4"
        FACILITY_CERTIFICATION_NUMBER = "F4"
        BATCH_NUMBER = "BT"
        REFERRAL_NUMBER = "9F"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2400AmtSegment(AmtSegment):
    """
    Loop 2400/Service Line Amounts
    """

    class AmountQualifierCode(str, Enum):
        """
        Code values for AMT01
        """

        TAX = "T"
        POSTAGE_CLAIMED = "F4"

    amount_qualifier_code: AmountQualifierCode


class Loop2400NteSegment(NteSegment):
    """
    Loop 2400/Claim Service Line Notes
    """

    class NoteReferenceCode(str, Enum):
        """
        NTE01 code values
        """

        ADDITIONAL_INFORMATION = "ADD"
        GOALS_REHAB_DISCHARGE_PLANS = "DCP"
        THIRD_PARTY_ORGANIZATION_NOTES = "TPO"

    note_reference_code: NoteReferenceCode


class Loop2410RefSegment(RefSegment):
    """
    Loop 2400/Claim Service Line Drug Identification Reference Identifier
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        LINK_SEQUENCE_NUMBER = "VY"
        PHARMACY_PRESCRIPTION_NUMBER = "XZ"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2420ANm1Segment(Nm1Segment):
    """
    Loop 2420A Rendering Provider Name
    """

    entity_identifier_code: Literal["82"]
    identification_code_qualifier: Literal["XX"]


class Loop2420APrvSegment(PrvSegment):
    """
    Loop 2420A Rendering Provider Speciality
    """

    provider_code: Literal["PE"]
    reference_identification_qualifier: Literal["PXC"]


class Loop2420ARefSegment(RefSegment):
    """
    Loop 2420A Rendering Provider Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2420BNm1Segment(Nm1Segment):
    """
    Loop 2420B Purchased Service Provider Name
    """

    entity_identifier_code: Literal["QB"]
    identification_code_qualifier: Literal["XX"]


class Loop2420BRefSegment(RefSegment):
    """
    Loop 2420B Purchased Service Provider Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2420CNm1Segment(Nm1Segment):
    """
    Loop 2420C Service Facility Location Name
    """

    entity_identifier_code: Literal["77"]


class Loop2420CRefSegment(RefSegment):
    """
    Loop 2420C Service Facility Location Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2420DNm1Segment(Nm1Segment):
    """
    Loop 2420D Supervising Provider Name
    """

    entity_identifier_code: Literal["DQ"]


class Loop2420DRefSegment(RefSegment):
    """
    Loop 2420D Supervising Provider Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2420ENm1Segment(Nm1Segment):
    """
    Loop 2420E Ordering Provider Name
    """

    entity_identifier_code: Literal["DK"]
    entity_type_qualifier: Literal["1"]


class Loop2420ERefSegment(RefSegment):
    """
    Loop 2420E Ordering Provider Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2420EPerSegment(PerSegment):
    """
    Loop 2420E Ordering Provider Contact Information
    """

    class CommunicationNumberQualifier(str, Enum):
        """
        Code values for PER03, PER05, PER07
        """

        ELECTRONIC_MAIL = "EM"
        TELEPHONE_EXTENSION = "EX"
        FACSIMILE = "FX"
        TELEPHONE = "TE"

    communication_number_qualifier_1: CommunicationNumberQualifier
    communication_number_qualifier_2: Optional[CommunicationNumberQualifier]
    communication_number_qualifier_3: Optional[CommunicationNumberQualifier]


class Loop2420FNm1Segment(Nm1Segment):
    """
    Loop 2420C Service Facility Location Name
    """

    class EntityIdentifierCode(str, Enum):
        """
        Code values for NM101
        """

        REFERRING_PROVIDER = "DN"
        PRIMARY_CARE_PROVIDER = "P3"

    entity_identifier_code: EntityIdentifierCode
    entity_type_qualifier: Literal["1"]


class Loop2420FRefSegment(RefSegment):
    """
    Loop 2420C Service Facility Location Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2420GNm1Segment(Nm1Segment):
    """
    Loop 2420G Ambulance Pickup Location
    """

    entity_identifier_code: Literal["PW"]
    entity_type_qualifier: Literal["2"]


class Loop2420HNm1Segment(Nm1Segment):
    """
    Loop 2420H Ambulance Drop-Off Location
    """

    entity_identifier_code: Literal["45"]
    entity_type_qualifier: Literal["2"]


class Loop2430DtpSegment(DtpSegment):
    """
    Loop 2430 Line Check or Remittance Date
    """

    date_time_qualifier: Literal["573"]
    date_time_period_format_qualifier: Literal["D8"]


class Loop2430AmtSegment(AmtSegment):
    """
    Loop 2430 Remaining Patient Liability
    """

    amount_qualifier_code: Literal["EAF"]



"""
test_segments.py

Tests the Pydantic X12 segment models
"""
from x12.segments import *
from decimal import Decimal
from x12.models import X12Delimiters


def test_aaa_segment():
    segment_data = {
        "response_code": "Y",
        "reject_reason_code": "42",
        "follow_up_action_code": "Y",
    }
    aaa_segment: AaaSegment = AaaSegment(**segment_data)
    assert aaa_segment.x12() == "AAA*Y**42*Y~"


def test_amt_segment():
    segment_data = {"amount_qualifier_code": "R", "monetary_amount": Decimal("37.5")}
    amt_segment: AmtSegment = AmtSegment(**segment_data)
    assert amt_segment.x12() == "AMT*R*37.50~"


def test_bht_segment():
    segment_data = {
        "hierarchical_structure_code": "0022",
        "transaction_set_purpose_code": "01",
        "submitter_transactional_identifier": "1c257041a955432091b0c073d788d29a",
        "transaction_set_creation_date": "19980101",
        "transaction_set_creation_time": "1400",
        "transaction_type_code": "RT",
    }

    bht_segment: BhtSegment = BhtSegment(**segment_data)
    assert (
        bht_segment.x12()
        == "BHT*0022*01*1c257041a955432091b0c073d788d29a*19980101*1400*RT~"
    )


def test_cas_segment():
    segment_data = {
        "adjustment_reason_code_1": "PR",
        "monetary_amount_1": 1.0,
        "quantity_1": 7.93,
    }

    cas_segment: CasSegment = CasSegment(**segment_data)
    assert cas_segment.x12() == "CAS*PR*1.00*7.93~"


def test_clm_segment():
    segment_data = {
        "patient_control_number": "26463774",
        "total_claim_charge_amount": "100.00",
        "health_care_service_location_information": "11:B:1",
        "provider_or_supplier_signature_indicator": "Y",
        "provider_accept_assignment_code": "A",
        "benefit_assignment_certification_indicator": "Y",
        "release_of_information_code": "I",
    }

    clm_segment: ClmSegment = ClmSegment(**segment_data)
    assert clm_segment.x12() == "CLM*26463774*100.00***11:B:1*Y*A*Y*I~"


def test_cn1_segment():
    segment_data = {"contract_type_code": "02", "contract_amount": "550"}

    cn1_segment: Cn1Segment = Cn1Segment(**segment_data)
    assert cn1_segment.x12() == "CN1*02*550.00~"


def test_crc_segment():
    segment_data = {
        "code_category": "E1",
        "certification_condition_indicator": "Y",
        "condition_code_1": "L1",
    }
    crc_segment: CrcSegment = CrcSegment(**segment_data)
    assert crc_segment.x12() == "CRC*E1*Y*L1~"


def test_cr1_segment():
    segment_data = {
        "weight_measurement_code": "LB",
        "patient_weight": "140",
        "ambulance_transport_reason_code": "A",
        "mileage_measurement_code": "DH",
        "transport_distance": "12.0",
        "stretcher_purpose_description": "UNCONSCIOUS",
    }

    cr1_segment: Cr1Segment = Cr1Segment(**segment_data)
    assert cr1_segment.x12() == "CR1*LB*140.00**A*DH*12.00****UNCONSCIOUS~"


def test_cr2_segment():
    segment_data = {"patient_condition_code": "M"}
    cr2_segment: Cr2Segment = Cr2Segment(**segment_data)
    assert cr2_segment.x12() == "CR2********M~"


def test_cr3_segment():
    segment_data = {
        "certification_type_code": "I",
        "unit_basis_measurement_code": "MO",
        "durable_medical_equipment_duration": "6.00",
    }
    cr3_segment: Cr3Segment = Cr3Segment(**segment_data)
    assert cr3_segment.x12() == "CR3*I*MO*6.00~"


def test_ctp_segment():
    segment_data = {"quantity": "2.00", "composite_unit_of_measure": "UN"}
    ctp_segment: CtpSegment = CtpSegment(**segment_data)
    assert ctp_segment.x12() == "CTP****2.00*UN~"


def test_cur_segment():
    segment_data = {"entity_identifier_code": "85", "currency_code": "USD"}

    cur_segment: CurSegment = CurSegment(**segment_data)
    assert cur_segment.x12() == "CUR*85*USD~"


def test_dmg_segment():
    segment_data = {
        "date_time_period_format_qualifier": "D8",
        "date_time_period": "19430917",
        "gender_code": "M",
    }

    dmg_segment: DmgSegment = DmgSegment(**segment_data)
    assert dmg_segment.x12() == "DMG*D8*19430917*M~"


def test_dtp_segment_date_period():
    segment_data = {
        "date_time_qualifier": "291",
        "date_time_period_format_qualifier": "D8",
        "date_time_period": "20051015",
    }

    dtp_segment: DtpSegment = DtpSegment(**segment_data)
    assert dtp_segment.x12() == "DTP*291*D8*20051015~"


def test_dtp_segment_date_range():
    segment_data = {
        "date_time_qualifier": "291",
        "date_time_period_format_qualifier": "RD8",
        "date_time_period": "20051015-20051201",
    }

    dtp_segment: DtpSegment = DtpSegment(**segment_data)
    assert dtp_segment.x12() == "DTP*291*RD8*20051015-20051201~"


def test_eb_segment():
    segment_data = {
        "eligibility_benefit_information": "B",
        "service_type_code": [
            "1",
            "33",
            "35",
            "47",
            "86",
            "88",
            "98",
            "AL",
            "MH",
            "UC",
        ],
        "insurance_type_code": "HM",
        "plan_coverage_description": "GOLD 123 PLAN",
        "time_period_qualifier": "27",
        "benefit_amount": "10.00",
        "inplan_network_indicator": "Y",
    }

    eb_segment: EbSegment = EbSegment(**segment_data)
    assert (
        eb_segment.x12()
        == "EB*B**1^33^35^47^86^88^98^AL^MH^UC*HM*GOLD 123 PLAN*27*10.00*****Y~"
    )


def test_eq_segment():
    segment_data = {
        "service_type_code": ["98", "34", "44", "81", "A0", "A3"],
        "coverage_level_code": "FAM",
    }

    eq_segment: EqSegment = EqSegment(**segment_data)
    assert eq_segment.x12() == "EQ*98^34^44^81^A0^A3**FAM~"


def test_frm_segment():
    segment_data = {"question_number_letter": "12", "question_response_1": "N"}

    frm_segment: FrmSegment = FrmSegment(**segment_data)
    assert frm_segment.x12() == "FRM*12*N~"


def test_ge_segment():
    segment_data = {
        "number_of_transaction_sets_included": 1,
        "group_control_number": "1",
    }

    ge_segment: GeSegment = GeSegment(**segment_data)
    assert ge_segment.x12() == "GE*1*1~"


def test_gs_segment():
    segment_data = {
        "functional_identifier_code": "HS",
        "application_sender_code": "000000005",
        "application_receiver_code": "54321",
        "functional_group_creation_date": "20131031",
        "functional_group_creation_time": "1147",
        "group_control_number": "1",
        "responsible_agency_code": "X",
        "version_identifier_code": "005010X279A",
    }

    gs_segment: GsSegment = GsSegment(**segment_data)
    assert gs_segment.x12() == "GS*HS*000000005*54321*20131031*1147*1*X*005010X279A~"


def test_hcp_segment():
    segment_data = {
        "pricing_methodology": "03",
        "repriced_allowed_amount": "100.00",
        "repriced_saving_amount": "10.00",
        "repricing_organization_identifier": "RPO12345",
    }
    hcp_segment: HcpSegment = HcpSegment(**segment_data)
    assert hcp_segment.x12() == "HCP*03*100.00*10.00*RPO12345~"


def test_hi_segment():
    segment_data = {
        "health_care_code_1": ["BK", "8901"],
        "health_care_code_2": ["BF", "87200"],
        "health_care_code_3": ["BF", "5559"],
    }
    hi_segment: HiSegment = HiSegment(**segment_data)
    assert hi_segment.x12() == "HI*BK:8901*BF:87200*BF:5559~"


def test_hl_segment():
    segment_data = {
        "hierarchical_id_number": "3",
        "hierarchical_parent_id_number": "2",
        "hierarchical_level_code": "22",
        "hierarchical_child_code": "0",
    }

    hl_segment: HlSegment = HlSegment(**segment_data)
    assert hl_segment.x12() == "HL*3*2*22*0~"


def test_hsd_segment():
    segment_data = {
        "quantity_qualifier": "VS",
        "quantity": 12.00,
        "measurement_code": "WK",
        "sample_selection_modulus": 3.00,
        "time_period_qualifier": "34",
        "period_count": 1.00,
    }
    hsd_segment: HsdSegment = HsdSegment(**segment_data)
    assert hsd_segment.x12() == "HSD*VS*12.00*WK*3.00*34*1.00~"


def test_iea_segment():
    segment_data = {
        "number_of_included_functional_groups": 1,
        "interchange_control_number": "000000907",
    }

    iea_segment: IeaSegment = IeaSegment(**segment_data)
    assert iea_segment.x12() == "IEA*1*000000907~"


def test_iii_segment():
    segment_data = {"code_list_qualifier_code": "ZZ", "industry_code": "21"}
    iii_segment: IiiSegment = IiiSegment(**segment_data)
    assert iii_segment.x12() == "III*ZZ*21~"


def test_ins_segment():
    segment_data = {
        "member_indicator": "Y",
        "individual_relationship_code": "18",
        "maintenance_type_code": "021",
        "maintenance_reason_code": "28",
        "benefit_status_code": "A",
        "employment_status_code": "FT",
    }
    ins_segment: InsSegment = InsSegment(**segment_data)
    assert ins_segment.x12() == "INS*Y*18*021*28*A***FT~"


def test_isa_segment():
    segment_data = {
        "authorization_information_qualifier": "03",
        "authorization_information": "9876543210",
        "security_information_qualifier": "01",
        "security_information": "9876543210",
        "interchange_sender_qualifier": "30",
        "interchange_sender_id": "000000005      ",
        "interchange_receiver_qualifier": "30",
        "interchange_receiver_id": "12345          ",
        "interchange_date": "131031",
        "interchange_time": "1147",
        "repetition_separator": "^",
        "interchange_control_version_number": "00501",
        "interchange_control_number": "000000907",
        "acknowledgment_requested": "1",
        "interchange_usage_indicator": "T",
        "component_element_separator": ":",
    }

    isa_segment: IsaSegment = IsaSegment(**segment_data)
    assert (
        isa_segment.x12()
        == "ISA*03*9876543210*01*9876543210*30*000000005      *30*12345          *131031*1147*^*00501*000000907*1*T*:~"
    )


def test_k3_segment():
    segment_data = {"fixed_format_information": "STATE DATA REQUIREMENT"}
    k3_segment: K3Segment = K3Segment(**segment_data)
    assert k3_segment.x12() == "K3*STATE DATA REQUIREMENT~"


def test_le_segment():
    segment_data = {"loop_id_code": "2120"}
    le_segment: LeSegment = LeSegment(**segment_data)
    assert le_segment.x12() == "LE*2120~"


def test_lin_segment():
    segment_data = {
        "product_service_id_qualifier": "N4",
        "national_drug_code_universal_product_number": "01234567891",
    }
    lin_segment: LinSegment = LinSegment(**segment_data)
    assert lin_segment.x12() == "LIN**N4*01234567891~"


def test_lq_segment():
    segment_data = {"code_list_qualifier_code": "UT", "form_identifier": "1.02"}
    lq_segment: LqSegment = LqSegment(**segment_data)
    assert lq_segment.x12() == "LQ*UT*1.02~"


def test_ls_segment():
    segment_data = {"loop_id_code": "2120"}
    ls_segment: LsSegment = LsSegment(**segment_data)
    assert ls_segment.x12() == "LS*2120~"


def test_lx_segment():
    segment_data = {"assigned_number": 1}
    lx_segment: LxSegment = LxSegment(**segment_data)
    assert lx_segment.x12() == "LX*1~"


def test_mea_segment():
    segment_data = {
        "measurement_reference_id_code": "TR",
        "measurement_qualifier": "R1",
        "measurement_value": "113.40",
    }
    mea_segment: MeaSegment = MeaSegment(**segment_data)
    assert mea_segment.x12() == "MEA*TR*R1*113.40~"


def test_moa_segment():
    segment_data = {"claim_payment_remark_code_1": "A4"}
    moa_segment: MoaSegment = MoaSegment(**segment_data)
    assert moa_segment.x12() == "MOA***A4~"


def test_msg_segment():
    segment_data = {"free_form_text": "this is free form text"}
    msg_segment: MsgSegment = MsgSegment(**segment_data)
    assert msg_segment.x12() == "MSG*this is free form text~"


def test_n3_segment():
    segment_data = {
        "address_information_1": "201 PARK AVENUE",
        "address_information_2": "SUITE 300",
    }
    n3_segment: N3Segment = N3Segment(**segment_data)
    assert n3_segment.x12() == "N3*201 PARK AVENUE*SUITE 300~"


def test_n4_segment():
    segment_data = {
        "city_name": "KANSAS CITY",
        "state_province_code": "MO",
        "postal_code": "64108",
    }
    n4_segment: N4Segment = N4Segment(**segment_data)
    assert n4_segment.x12() == "N4*KANSAS CITY*MO*64108~"


def test_nm1_segment():
    segment_data = {
        "entity_identifier_code": "PR",
        "entity_type_qualifier": "2",
        "name_last_or_organization_name": "ACME",
        "identification_code_qualifier": "PI",
        "identification_code": "12345",
    }
    nm1_segment: Nm1Segment = Nm1Segment(**segment_data)
    assert nm1_segment.x12() == "NM1*PR*2*ACME*****PI*12345~"


def test_nte_segment():
    segment_data = {
        "note_reference_code": "ADD",
        "description": "SURGERY WAS UNUSUALLY LONG BECAUSE [FILL IN REASON]",
    }
    nte_segment: NteSegment = NteSegment(**segment_data)
    assert (
        nte_segment.x12()
        == "NTE*ADD*SURGERY WAS UNUSUALLY LONG BECAUSE [FILL IN REASON]~"
    )


def test_oi_segment():
    segment_data = {
        "benefits_assignment_certification": "Y",
        "patient_signature_source_code": "P",
        "release_of_information_code": "Y",
    }

    oi_segment: OiSegment = OiSegment(**segment_data)
    assert oi_segment.x12() == "OI***Y*P**Y~"


def test_pat_segment():
    segment_data = {
        "unit_basis_measurement_code": "01",
        "patient_weight": "146.00",
    }
    pat_segment: PatSegment = PatSegment(**segment_data)
    assert pat_segment.x12() == "PAT******01*146.00~"


def test_prv_segment():
    segment_data = {
        "provider_code": "RF",
        "reference_identification_qualifier": "PXC",
        "reference_identification": "207Q00000X",
    }
    prv_segment: PrvSegment = PrvSegment(**segment_data)
    assert prv_segment.x12() == "PRV*RF*PXC*207Q00000X~"


def test_ps1_segment():
    segment_data = {
        "purchased_service_provider_identifier": "PN222222",
        "purchased_service_charge_amount": "110.00",
    }
    ps1_segment: Ps1Segment = Ps1Segment(**segment_data)
    assert ps1_segment.x12() == "PS1*PN222222*110.00~"


def test_pwk_segment():
    segment_data = {
        "segment_name": "PWK",
        "report_type_code": "OZ",
        "report_transmission_code": "BM",
        "identification_code_qualifier": "AC",
        "identification_code": "DMN0012",
    }

    pwk_segment: PwkSegment = PwkSegment(**segment_data)
    assert pwk_segment.x12() == "PWK*OZ*BM***AC*DMN0012~"


def test_qty_segment():
    segment_data = {"quantity_qualifier": "PT", "quantity": "2.00"}
    qty_segment: QtySegment = QtySegment(**segment_data)
    assert qty_segment.x12() == "QTY*PT*2.00~"


def test_ref_segment():
    segment_data = {
        "reference_identification_qualifier": "EO",
        "reference_identification": "477563928",
    }
    ref_segment: RefSegment = RefSegment(**segment_data)
    assert ref_segment.x12() == "REF*EO*477563928~"


def test_sbr_segment():
    segment_data = {
        "payer_responsibility_code": "P",
        "group_policy_number": "2222-SJ",
        "claim_filing_indicator_code": "CI",
    }
    sbr_segment: SbrSegment = SbrSegment(**segment_data)
    assert sbr_segment.x12() == "SBR*P**2222-SJ******CI~"


def test_se_segment():
    segment_data = {
        "transaction_segment_count": 17,
        "transaction_set_control_number": "0001",
    }

    se_segment: SeSegment = SeSegment(**segment_data)
    assert se_segment.x12() == "SE*17*0001~"


def test_st_segment():
    segment_data = {
        "transaction_set_identifier_code": "270",
        "transaction_set_control_number": "0001",
        "implementation_convention_reference": "005010X279A1",
    }

    st_segment: StSegment = StSegment(**segment_data)
    assert st_segment.x12() == "ST*270*0001*005010X279A1~"


def test_sv1_segment():
    segment_data = {
        "product_service_id_qualifier": "HC:99213",
        "line_item_charge_amount": "40.00",
        "unit_basis_measurement_code": "UN",
        "service_unit_count": "1.00",
        "composite_diagnosis_code_pointer": "1",
    }
    sv1_segment: Sv1Segment = Sv1Segment(**segment_data)
    assert sv1_segment.x12() == "SV1*HC:99213*40.00*UN*1.00***1~"


def test_sv5_segment():
    segment_data = {
        "product_service_id_qualifier": "HC:A4631",
        "unit_basis_measurement_code": "DA",
        "length_of_medical_necessity": "30.00",
        "dme_rental_price": "50.00",
        "dme_purchase_price": "5000.00",
        "rental_unit_price_indicator": "4",
    }
    sv5_segment: Sv5Segment = Sv5Segment(**segment_data)
    assert sv5_segment.x12() == "SV5*HC:A4631*DA*30.00*50.00*5000.00*4~"


def test_svd_segment():
    segment_data = {
        "other_payer_primary_identifier": "43",
        "service_line_paid_amount": "55.00",
        "composite_medical_procedure_identifier": "HC:84550",
        "paid_service_count": "3.00",
    }
    svd_segment: SvdSegment = SvdSegment(**segment_data)
    assert svd_segment.x12() == "SVD*43*55.00*HC:84550**3.00~"


def test_trn_segment():
    segment_data = {
        "trace_type_code": "1",
        "reference_identification_1": "98175-012547",
        "originating_company_identifier": "8877281234",
        "reference_identification_2": "RADIOLOGY",
    }

    trn_segment: TrnSegment = TrnSegment(**segment_data)
    assert trn_segment.x12() == "TRN*1*98175-012547*8877281234*RADIOLOGY~"


def test_segment_lookup():
    """Basic validations to ensure that the segment lookup loads"""

    assert len(SEGMENT_LOOKUP) > 0
    assert "GE" in SEGMENT_LOOKUP
    assert "GS" in SEGMENT_LOOKUP
    assert "IEA" in SEGMENT_LOOKUP
    assert "ISA" in SEGMENT_LOOKUP
    assert "ST" in SEGMENT_LOOKUP
    assert "SE" in SEGMENT_LOOKUP


def test_x12_with_custom_delimiters():
    """tests x12 generation where custom delimiters are used"""
    x12_delimiters: X12Delimiters = X12Delimiters(
        element_separator="|", segment_terminator="?"
    )

    segment_data = {
        "delimiters": x12_delimiters.dict(),
        "trace_type_code": "1",
        "reference_identification_1": "98175-012547",
        "originating_company_identifier": "8877281234",
        "reference_identification_2": "RADIOLOGY",
    }

    trn_segment: TrnSegment = TrnSegment(**segment_data)
    assert trn_segment.x12() == "TRN|1|98175-012547|8877281234|RADIOLOGY?"
